from os import path
import posixpath
from silx.io.url import DataUrl
from silx.io.dictdump import h5todict
from ..utils import check_supported
from ..io.utils import get_first_hdf5_entry, get_h5_value
from ..io.writer import NXProcessWriter
from ..misc.histogram import PartialHistogram, VolumeHistogram, hist_as_2Darray
from ..misc.histogram_cuda import CudaVolumeHistogram
from ..resources.logger import Logger, LoggerOrPrint
from .utils import parse_params_values
from .cli_configs import HistogramConfig


class VolumesHistogram:
    """
    A class for extracting or computing histograms of one or several volumes.
    """

    available_backends = {
        "numpy": VolumeHistogram,
        "cuda": CudaVolumeHistogram,
    }

    def __init__(
        self, fnames, output_file, chunk_size_slices=100, chunk_size_GB=None, nbins=1e6, logger=None, backend="cuda"
    ):
        """
        Initialize a VolumesHistogram object.

        Parameters
        -----------
        fnames: list of str
            List of paths to HDF5 files.
            To specify an entry for each file name, use the "?" separator:
            /path/to/file.h5?entry0001
        output_file: str
            Path to the output file
        write_histogram_if_computed: bool, optional
            Whether to write histograms that are computed to a file.
            Some volumes might be missing their histogram. In this case, the histogram
            is computed, and the result is written to a dedicated file in the same
            directory as 'output_file'.
            Default is True.
        """
        self._get_files_and_entries(fnames)
        self.chunk_size_slices = chunk_size_slices
        self.chunk_size_GB = chunk_size_GB
        self.nbins = nbins
        self.logger = LoggerOrPrint(logger)
        self.output_file = output_file
        self._get_histogrammer_backend(backend)

    def _get_files_and_entries(self, fnames):
        res_fnames = []
        res_entries = []
        for fname in fnames:
            if "?" not in fname:
                entry = None
            else:
                fname, entry = fname.split("?")
                if entry == "":
                    entry = None
            res_fnames.append(fname)
            res_entries.append(entry)
        self.fnames = res_fnames
        self.entries = res_entries

    def _get_histogrammer_backend(self, backend):
        check_supported(backend, self.available_backends.keys(), "histogram backend")
        self.VolumeHistogramClass = self.available_backends[backend]

    def _get_config_onevolume(self, fname, entry, data_shape):
        return {
            "chunk_size_slices": self.chunk_size_slices,
            "chunk_size_GB": self.chunk_size_GB,
            "bins": self.nbins,
            "filename": fname,
            "entry": entry,
            "volume_shape": data_shape,
        }

    def _get_config(self):
        conf = self._get_config_onevolume("", "", None)
        conf.pop("filename")
        conf.pop("entry")
        conf["filenames"] = self.fnames
        conf["entries"] = [entry if entry is not None else "None" for entry in self.entries]
        return conf

    def _write_histogram_onevolume(self, fname, entry, histogram, data_shape):
        output_file = (
            path.join(path.dirname(self.output_file), path.splitext(path.basename(fname))[0])
            + "_histogram"
            + path.splitext(fname)[1]
        )
        self.logger.info("Writing histogram of %s into %s" % (fname, output_file))
        writer = NXProcessWriter(output_file, entry, filemode="w", overwrite=True)
        writer.write(
            hist_as_2Darray(histogram), "histogram", config=self._get_config_onevolume(fname, entry, data_shape)
        )

    def get_histogram_single_volume(self, fname, entry, write_histogram_if_computed=True, return_config=False):
        entry = entry or get_first_hdf5_entry(fname)
        hist_path = posixpath.join(entry, "histogram", "results", "data")
        hist_cfg_path = posixpath.join(entry, "histogram", "configuration")
        rec_path = posixpath.join(entry, "reconstruction", "results", "data")
        rec_url = DataUrl(file_path=fname, data_path=rec_path)
        hist = get_h5_value(fname, hist_path)
        config = None
        if hist is None:
            self.logger.info("No histogram found in %s, computing it" % fname)
            vol_histogrammer = self.VolumeHistogramClass(
                rec_url,
                chunk_size_slices=self.chunk_size_slices,
                chunk_size_GB=self.chunk_size_GB,
                nbins=self.nbins,
                logger=self.logger,
            )
            hist = vol_histogrammer.compute_volume_histogram()
            if write_histogram_if_computed:
                self._write_histogram_onevolume(fname, entry, hist, vol_histogrammer.data_shape)
            else:
                if return_config:
                    raise ValueError(
                        "return_config must be set to True to get configuration for non-existing histograms"
                    )
            hist = hist_as_2Darray(hist)
        config = h5todict(path.splitext(fname)[0] + "_histogram" + path.splitext(fname)[1], path=hist_cfg_path)
        if return_config:
            return hist, config
        else:
            return hist

    def get_histogram(self, return_config=False):
        histograms = []
        configs = []
        for fname, entry in zip(self.fnames, self.entries):
            self.logger.info("Getting histogram for %s" % fname)
            hist, conf = self.get_histogram_single_volume(fname, entry, return_config=True)
            histograms.append(hist)
            configs.append(conf)
        self.logger.info("Merging histograms")
        histogrammer = PartialHistogram(method="fixed_bins_number", num_bins=self.nbins)
        hist = histogrammer.merge_histograms(histograms, dont_truncate_bins=True)
        if return_config:
            return hist, configs
        else:
            return hist

    def merge_histograms_configurations(self, configs):
        if configs is None or len(configs) == 0:
            return
        res_config = {"volume_shape": list(configs[0]["volume_shape"])}
        res_config["volume_shape"][0] = 0
        for conf in configs:
            nz, ny, nx = conf["volume_shape"]
            res_config["volume_shape"][0] += nz
        res_config["volume_shape"] = tuple(res_config["volume_shape"])
        return res_config

    def write_histogram(self, hist, config=None):
        self.logger.info("Writing final histogram to %s" % (self.output_file))
        config = config or {}
        base_config = self._get_config()
        base_config.pop("volume_shape")
        config.update(base_config)
        writer = NXProcessWriter(self.output_file, "entry0000", filemode="w", overwrite=True)
        writer.write(hist_as_2Darray(hist), "histogram", config=config)


def histogram_cli():
    args = parse_params_values(HistogramConfig, parser_description="Extract/compute histogram of volume(s).")
    logger = Logger("nabu_histogram", level=args["loglevel"], logfile="nabu_histogram.log")
    output = args["output_file"].split("?")[0]
    if path.exists(output):
        logger.fatal("Output file %s already exists, not overwriting it" % output)
        exit(1)
    chunk_size_gb = float(args["chunk_size_GB"])
    if chunk_size_gb <= 0:
        chunk_size_gb = None
    histogramer = VolumesHistogram(
        args["h5_file"],
        output,
        chunk_size_slices=int(args["chunk_size_slices"]),
        chunk_size_GB=chunk_size_gb,
        nbins=int(args["bins"]),
        logger=logger,
    )
    hist, configs = histogramer.get_histogram(return_config=True)
    config = histogramer.merge_histograms_configurations(configs)
    histogramer.write_histogram(hist, config=config)


if __name__ == "__main__":
    histogram_cli()



# Backward compat.
from ..estimation.alignment import AlignmentBase
from ..estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationGrowingWindow,
    CenterOfRotationSlidingWindow,
)
from ..estimation.translation import DetectorTranslationAlongBeam
from ..estimation.focus import CameraFocus
from ..estimation.tilt import CameraTilt



from nabu.utils import Progress
from .cli_configs import StitchingConfig
from ..pipeline.config import parse_nabu_config_file
from nabu.stitching.z_stitching import z_stitching
from nabu.stitching.config import dict_to_config_obj
from .utils import parse_params_values


def main():
    args = parse_params_values(
        StitchingConfig,
        parser_description="Run stitching from a configuration file. Configuration can be obtain from `stitching-config`",
    )

    conf_dict = parse_nabu_config_file(args["input_file"], allow_no_value=True)

    stitching_config = dict_to_config_obj(conf_dict)

    progress = Progress("z-stitching")
    progress.set_name("initialize z-stitching")
    progress.set_advancement(0)
    z_stitching(stitching_config, progress=progress)
    exit(0)



""" This module is  meant to contain classes which are in the process of being superseed by new classes depending on recent packages 
    with fast development cycles in order to be able to  fall-back in two cases : 
             -- the new packages, or one of their parts,  break from one version to another.
             -- For parts of Nabu which need some extra time to adapt.    
"""
from ..resources.logger import LoggerOrPrint
from ..utils import check_supported
from ..io.writer import Writers, LegacyNXProcessWriter
from ..resources.utils import is_hdf5_extension
from os import path, mkdir
from .params import files_formats


class WriterConfigurator:
    """No dependency on  tomoscan for this class. The new class would be WriterManager which depend on tomoscan."""

    _overwrite_warned = False

    def __init__(
        self,
        output_dir,
        file_prefix,
        file_format="hdf5",
        overwrite=False,
        start_index=None,
        logger=None,
        nx_info=None,
        write_histogram=False,
        histogram_entry="entry",
        writer_options=None,
        extra_options=None,
    ):
        """
        Create a Writer from a set of parameters.

        Parameters
        ----------
        output_dir: str
            Directory where the file(s) will be written.
        file_prefix: str
            File prefix (without leading path)
        start_index: int, optional
            Index to start the files numbering (filename_0123.ext).
            Default is 0.
            Ignored for HDF5 extension.
        logger: nabu.resources.logger.Logger, optional
            Logger object
        nx_info: dict, optional
            Dictionary containing the nexus information.
        write_histogram: bool, optional
            Whether to also write a histogram of data. If set to True, it will configure
            an additional "writer".
        histogram_entry: str, optional
            Name of the HDF5 entry for the output histogram file, if write_histogram is True.
            Ignored if the output format is already HDF5 : in this case, nx_info["entry"] is taken.
        writer_options: dict, optional
            Other advanced options to pass to Writer class.
        """
        self.logger = LoggerOrPrint(logger)
        self.start_index = start_index
        self.write_histogram = write_histogram
        self.overwrite = overwrite
        writer_options = writer_options or {}
        self.extra_options = extra_options or {}

        check_supported(file_format, list(Writers.keys()), "output file format")

        self._set_output_dir(output_dir)
        self._set_file_name(file_prefix, file_format)

        # Init Writer
        writer_cls = Writers[file_format]
        writer_args = [self.fname]
        writer_kwargs = self._get_initial_writer_kwarg()
        self._writer_exec_args = []
        self._writer_exec_kwargs = {}
        self._is_hdf5_output = is_hdf5_extension(file_format)

        if self._is_hdf5_output:
            writer_kwargs["entry"] = nx_info["entry"]
            writer_kwargs["filemode"] = "a"
            writer_kwargs["overwrite"] = overwrite
            self._writer_exec_args.append(nx_info["process_name"])
            self._writer_exec_kwargs["processing_index"] = nx_info["processing_index"]
            self._writer_exec_kwargs["config"] = nx_info["config"]
        else:
            writer_kwargs["start_index"] = self.start_index
            if writer_options.get("tiff_single_file", False) and "tif" in file_format:
                do_append = writer_options.get("single_tiff_initialized", False)
                writer_kwargs.update({"multiframe": True, "append": do_append})

        if files_formats.get(file_format, None) == "jp2":
            cratios = self.extra_options.get("jpeg2000_compression_ratio", None)
            if cratios is not None:
                cratios = [cratios]
            writer_kwargs["cratios"] = cratios
            writer_kwargs["float_clip_values"] = self.extra_options.get("float_clip_values", None)
        self.writer = writer_cls(*writer_args, **writer_kwargs)

        if self.write_histogram and not (self._is_hdf5_output):
            self._init_separate_histogram_writer(histogram_entry)

    def _get_initial_writer_kwarg(self):
        return {}

    def _set_output_dir(self, output_dir):
        self.output_dir = output_dir
        if path.exists(self.output_dir):
            if not (path.isdir(self.output_dir)):
                raise ValueError(
                    "Unable to create directory %s: already exists and is not a directory" % self.output_dir
                )
        else:
            self.logger.debug("Creating directory %s" % self.output_dir)
            mkdir(self.output_dir)

    def _set_file_name(self, file_prefix, file_format):
        self.file_prefix = file_prefix
        self.file_format = file_format
        self.fname = path.join(self.output_dir, file_prefix + "." + file_format)
        if path.exists(self.fname):
            err = "File already exists: %s" % self.fname
            if self.overwrite:
                if not (WriterConfigurator._overwrite_warned):
                    self.logger.warning(err + ". It will be overwritten as requested in configuration")
                    WriterConfigurator._overwrite_warned = True
            else:
                self.logger.fatal(err)
                raise ValueError(err)

    def _init_separate_histogram_writer(self, hist_entry):
        hist_fname = path.join(self.output_dir, "histogram_%06d.hdf5" % self.start_index)
        self.histogram_writer = LegacyNXProcessWriter(
            hist_fname,
            entry=hist_entry,
            filemode="w",
            overwrite=True,
        )

    def get_histogram_writer(self):
        if not (self.write_histogram):
            return None
        if self._is_hdf5_output:
            return self.writer
        else:
            return self.histogram_writer

    def write_data(self, data):
        self.writer.write(data, *self._writer_exec_args, **self._writer_exec_kwargs)



import numpy as np
from .rotation import Rotation
from ..utils import get_cuda_srcfile, updiv
from ..cuda.utils import __has_pycuda__, copy_array
from ..cuda.processing import CudaProcessing

if __has_pycuda__:
    from ..cuda.kernel import CudaKernel
    import pycuda.gpuarray as garray
    import pycuda.driver as cuda


class CudaRotation(Rotation):
    def __init__(self, shape, angle, center=None, mode="edge", reshape=False, cuda_options=None, **sk_kwargs):
        if center is None:
            center = ((shape[1] - 1) / 2.0, (shape[0] - 1) / 2.0)
        super().__init__(shape, angle, center=center, mode=mode, reshape=reshape, **sk_kwargs)
        self._init_cuda_rotation(cuda_options)

    def _init_cuda_rotation(self, cuda_options):
        cuda_options = cuda_options or {}
        self.cuda_processing = CudaProcessing(**cuda_options)
        self._allocate_arrays()
        self._init_rotation_kernel()

    def _allocate_arrays(self):
        self._d_image_cua = cuda.np_to_array(np.zeros(self.shape, "f"), "C")
        self.cuda_processing.init_arrays_to_none(["d_output"])

    def _init_rotation_kernel(self):
        self.cuda_rotation_kernel = CudaKernel("rotate", get_cuda_srcfile("rotation.cu"))
        self.texref_image = self.cuda_rotation_kernel.module.get_texref("tex_image")
        self.texref_image.set_filter_mode(cuda.filter_mode.LINEAR)  # bilinear
        self.texref_image.set_address_mode(0, cuda.address_mode.CLAMP)  # TODO tune
        self.texref_image.set_address_mode(1, cuda.address_mode.CLAMP)  # TODO tune
        self.cuda_rotation_kernel.prepare("Piiffff", [self.texref_image])
        self.texref_image.set_array(self._d_image_cua)
        self._cos_theta = np.cos(np.deg2rad(self.angle))
        self._sin_theta = np.sin(np.deg2rad(self.angle))
        self._Nx = np.int32(self.shape[1])
        self._Ny = np.int32(self.shape[0])
        self._center_x = np.float32(self.center[0])
        self._center_y = np.float32(self.center[1])
        self._block = (32, 32, 1)  # tune ?
        self._grid = (updiv(self.shape[1], self._block[1]), updiv(self.shape[0], self._block[0]), 1)

    def rotate(self, img, output=None, do_checks=True):
        copy_array(self._d_image_cua, img, check=do_checks)
        if output is not None:
            d_out = output
        else:
            self.cuda_processing.allocate_array("d_output", self.shape, np.float32)
            d_out = self.cuda_processing.d_output
        self.cuda_rotation_kernel(
            d_out,
            self._Nx,
            self._Ny,
            self._cos_theta,
            self._sin_theta,
            self._center_x,
            self._center_y,
            grid=self._grid,
            block=self._block,
        )
        return d_out

    __call__ = rotate



from . import *
from .writer import TIFFWriter as StandardTIFFWriter
from os import path
from tifffile import TiffWriter
import numpy as np


class TIFFWriter(StandardTIFFWriter):  # pylint: disable=E0102
    def __init__(
        self,
        fname,
        multiframe=False,
        start_index=0,
        heights_above_stage_mm=None,
        filemode=None,
        append=False,
        big_tiff=None,
    ):
        """
        Tiff writer.

        Parameters
        -----------
        fname: str
            Path to the output file name
        multiframe: bool, optional
            Whether to write all data in one single file. Default is False.
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            (unless multiframe is set to True).
            In this case, the output is a series of files `filename_0000.tif`,
            `filename_0001.tif`, etc. This parameter is the starting index for
            file names.
            This option is ignored when multiframe is True.
        heights_above_stage_mm: None or a list of heights
            if this parameters is given, the file names will be indexed with the height
        filemode: str, optional
            DEPRECATED. Will be ignored. Please refer to 'append'
        append: bool, optional
            Whether to append data to the file rather than overwriting. Default is False.
        big_tiff: bool, optional
            Whether to write in "big tiff" format: https://www.awaresystems.be/imaging/tiff/bigtiff.html
            Default is True when multiframe is True.
            Note that default "standard" tiff cannot exceed 4 GB.

        Notes
        ------
        If multiframe is False (default), then each image will be written in a
        dedicated tiff file.
        """
        super().__init__(
            fname, multiframe=multiframe, start_index=start_index, filemode=filemode, append=append, big_tiff=big_tiff
        )
        self.heights_above_stage_mm = heights_above_stage_mm

    def _write_tiff(self, data, config=None, filename=None):
        # TODO metadata
        filename = filename or self.fname
        with TiffWriter(filename, bigtiff=self.big_tiff, append=self.append) as tif:
            tif.write(data)

    def write(self, data, *args, config=None, **kwargs):
        # Single image, or multiple image in the same file
        if self.multiframe:
            self._write_tiff(data, config=config)
        # Multiple image, one file per image
        else:
            if len(data.shape) == 2:
                data = np.array([data])
            dirname, rel_filename = path.split(self.fname)
            prefix, ext = path.splitext(rel_filename)
            for i in range(data.shape[0]):
                if self.heights_above_stage_mm is None:
                    curr_rel_filename = prefix + str("_%06d" % (self.start_index + i)) + ext
                else:
                    value_mm = self.heights_above_stage_mm[i]
                    if value_mm < 0:
                        sign = "-"
                        value_mm = -value_mm
                    else:
                        sign = ""

                    part_mm = int(value_mm)
                    rest_um = (value_mm - part_mm) * 1000
                    part_um = int(rest_um)
                    rest_nm = (rest_um - part_um) * 1000
                    part_nm = int(rest_nm)

                    curr_rel_filename = prefix + "_{}{:06d}p{:03d}{:03d}".format(sign, part_mm, part_um, part_nm) + ext

                fname = path.join(dirname, curr_rel_filename)

                self._write_tiff(data[i], filename=fname, config=None)

    def get_filename(self):
        if self.multiframe:
            return self.fname
        else:
            return path.dirname(self.fname)



import os
import numpy as np
import pytest
import h5py
from nabu.testutils import utilstest, __do_long_tests__
from nabu.estimation.focus import CameraFocus


@pytest.fixture(scope="class")
def bootstrap_fcs(request):
    cls = request.cls
    cls.abs_tol_dist = 1e-2
    cls.abs_tol_tilt = 2.5e-4

    (
        cls.data,
        cls.img_pos,
        cls.pixel_size,
        (calib_data_std, calib_data_angle),
    ) = get_focus_data("test_alignment_focus.h5")
    (
        cls.angle_best_ind,
        cls.angle_best_pos,
        cls.angle_tilt_v,
        cls.angle_tilt_h,
    ) = calib_data_angle
    cls.std_best_ind, cls.std_best_pos = calib_data_std


def get_focus_data(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    with h5py.File(dataset_downloaded_path, "r") as hf:
        data = hf["/entry/instrument/detector/data"][()]
        img_pos = hf["/entry/instrument/detector/distance"][()]

        pixel_size = np.mean(
            [
                hf["/entry/instrument/detector/x_pixel_size"][()],
                hf["/entry/instrument/detector/y_pixel_size"][()],
            ]
        )

        angle_best_ind = hf["/calibration/focus/angle/best_img"][()]
        angle_best_pos = hf["/calibration/focus/angle/best_pos"][()]
        angle_tilt_v = hf["/calibration/focus/angle/tilt_v_rad"][()]
        angle_tilt_h = hf["/calibration/focus/angle/tilt_h_rad"][()]

        std_best_ind = hf["/calibration/focus/std/best_img"][()]
        std_best_pos = hf["/calibration/focus/std/best_pos"][()]

    calib_data_angle = (angle_best_ind, angle_best_pos, angle_tilt_v, angle_tilt_h)
    calib_data_std = (std_best_ind, std_best_pos)
    return data, img_pos, pixel_size, (calib_data_std, calib_data_angle)


@pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
@pytest.mark.usefixtures("bootstrap_fcs")
class TestFocus:
    def test_find_distance(self):
        focus_calc = CameraFocus()
        focus_pos, focus_ind = focus_calc.find_distance(self.data, self.img_pos)

        message = (
            "Computed focus motor position %f " % focus_pos + " and expected %f do not coincide" % self.std_best_pos
        )
        assert np.isclose(self.std_best_pos, focus_pos, atol=self.abs_tol_dist), message

        message = "Computed focus image index %f " % focus_ind + " and expected %f do not coincide" % self.std_best_ind
        assert np.isclose(self.std_best_ind, focus_ind, atol=self.abs_tol_dist), message

    def test_find_scintillator_tilt(self):
        focus_calc = CameraFocus()
        focus_pos, focus_ind, tilts_vh = focus_calc.find_scintillator_tilt(self.data, self.img_pos)

        message = (
            "Computed focus motor position %f " % focus_pos + " and expected %f do not coincide" % self.angle_best_pos
        )
        assert np.isclose(self.angle_best_pos, focus_pos, atol=self.abs_tol_dist), message

        message = (
            "Computed focus image index %f " % focus_ind + " and expected %f do not coincide" % self.angle_best_ind
        )
        assert np.isclose(self.angle_best_ind, focus_ind, atol=self.abs_tol_dist), message

        expected_tilts_vh = np.squeeze(np.array([self.angle_tilt_v, self.angle_tilt_h]))
        computed_tilts_vh = -tilts_vh / (self.pixel_size / 1000)
        message = "Computed tilts %s and expected %s do not coincide" % (
            computed_tilts_vh,
            expected_tilts_vh,
        )
        assert np.all(np.isclose(computed_tilts_vh, expected_tilts_vh, atol=self.abs_tol_tilt)), message

    def test_size_determination(self):
        inp_shape = [2162, 2560]
        exp_shape = np.array([2160, 2160])
        new_shape = CameraFocus()._check_img_block_size(inp_shape, 4, suggest_new_shape=True)

        message = "New suggested shape: %s and expected: %s do not coincide" % (new_shape, exp_shape)
        assert np.all(new_shape == exp_shape), message



from itertools import product
import posixpath
import tarfile
import os
import numpy as np
from silx.resources import ExternalResources
from silx.io.dictdump import dicttoh5, nxtodict, dicttonx
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from .io.utils import get_compacted_dataslices

utilstest = ExternalResources(
    project="nabu", url_base="http://www.silx.org/pub/nabu/data/", env_key="NABU_DATA", timeout=60
)

__big_testdata_dir__ = os.environ.get("NABU_BIGDATA_DIR")
if __big_testdata_dir__ is None or not (os.path.isdir(__big_testdata_dir__)):
    __big_testdata_dir__ = None

__do_long_tests__ = os.environ.get("NABU_LONG_TESTS", False)
if __do_long_tests__:
    try:
        __do_long_tests__ = bool(int(__do_long_tests__))
    except:
        __do_long_tests__ = False

__do_large_mem_tests__ = os.environ.get("NABU_LARGE_MEM_TESTS", False)
if __do_large_mem_tests__:
    try:
        __do_large_mem_tests__ = bool(int(__do_large_mem_tests__))
    except:
        __do_large_mem_tests__ = False


def generate_tests_scenarios(configurations):
    """
    Generate "scenarios" of tests.

    The parameter is a dictionary where:
      - the key is the name of a parameter
      - the value is a list of possible parameters

    This function returns a list of dictionary where:
      - the key is the name of a parameter
      - the value is one value of this parameter
    """
    scenarios = [{key: val for key, val in zip(configurations.keys(), p_)} for p_ in product(*configurations.values())]
    return scenarios


def get_data(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    return np.load(dataset_downloaded_path)


def get_big_data(filename):
    if __big_testdata_dir__ is None:
        return None
    return np.load(os.path.join(__big_testdata_dir__, filename))


def uncompress_file(compressed_file_path, target_directory):
    with tarfile.open(compressed_file_path) as f:
        f.extractall(path=target_directory)


def get_file(fname):
    downloaded_file = dataset_downloaded_path = utilstest.getfile(fname)
    if ".tar" in fname:
        uncompress_file(downloaded_file, os.path.dirname(downloaded_file))
        downloaded_file = downloaded_file.split(".tar")[0]
    return downloaded_file


def compare_arrays(arr1, arr2, tol, diff=None, absolute_value=True, percent=None, method="max", return_residual=False):
    """
    Utility to compare two arrays.

    Parameters
    ----------
    arr1: numpy.ndarray
        First array to compare
    arr2: numpy.ndarray
        Second array to compare
    tol: float
        Tolerance indicating whether arrays are close to eachother.
    diff: numpy.ndarray, optional
        Difference `arr1 - arr2`. If provided, this array is taken instead of `arr1`
        and `arr2`.
    absolute_value: bool, optional
        Whether to take absolute value of the difference.
    percent: float
        If set, a "relative" comparison is performed instead of a subtraction:
        `red(|arr1 - arr2|) / (red(|arr1|) * percent) < tol`
        where "red" is the reduction method (mean, max or median).
    method:
        Reduction method. Can be "max", "mean", or "median".

    Returns
    --------
    (is_close, residual) if return_residual is set to True
    is_close otherwise

    Examples
    --------
    When using method="mean" and absolute_value=True, this function computes
    the Mean Absolute Difference (MAD) metric.
    When also using percent=1.0, this computes the Relative Mean Absolute Difference
    (RMD) metric.
    """
    reductions = {
        "max": np.max,
        "mean": np.mean,
        "median": np.median,
    }
    if method not in reductions:
        raise ValueError("reduction method should be in %s" % str(list(reductions.keys())))
    if diff is None:
        diff = arr1 - arr2
    if absolute_value is not None:
        diff = np.abs(diff)
    residual = reductions[method](diff)
    if percent is not None:
        a1 = np.abs(arr1) if absolute_value else arr1
        residual /= reductions[method](a1)

    res = residual < tol
    if return_residual:
        res = res, residual
    return res


class SimpleHDF5TomoScanMock:
    def __init__(self, image_key):
        self._image_key = image_key

    @property
    def image_key(self):
        return self._image_key

    @image_key.setter
    def image_key(self, image_key):
        self._image_key = image_key

    def save_reduced_flats(self, *args, **kwargs):
        pass

    def save_reduced_darks(self, *args, **kwargs):
        pass


class NXDatasetMock:
    """
    An alternative to tomoscan.esrf.mock.MockHDF5, with a different interface.
    Attributes are not supported !
    """

    def __init__(self, data_volume, image_keys, rotation_angles=None, incident_energy=19.0, other_params=None):
        self.data_volume = data_volume
        self.n_proj = data_volume.shape[0]
        self.image_key = image_keys
        if rotation_angles is None:
            rotation_angles = np.linspace(0, 180, self.n_proj, False)
        self.rotation_angle = rotation_angles
        self.incident_energy = incident_energy
        assert image_keys.size == self.n_proj
        self._finalize_init(other_params)
        self.dataset_dict = None
        self.fname = None
        # Mocks more attributes
        self.dataset_scanner = SimpleHDF5TomoScanMock(image_key=self.image_key)
        self.kind = "hdf5"

    def _finalize_init(self, other_params):
        if other_params is None:
            other_params = {}
        default_params = {
            "detector": {
                "count_time": 0.05 * np.ones(self.n_proj, dtype="f"),
                "distance": 0.5,
                "field_of_view": "Full",
                "image_key_control": np.copy(self.image_key),
                "x_pixel_size": 6.5e-6,
                "y_pixel_size": 6.5e-6,
                "x_magnified_pixel_size": 6.5e-5,
                "y_magnified_pixel_size": 6.5e-5,
            },
            "sample": {
                "name": "dummy sample",
                "x_translation": 5e-4 * np.ones(self.n_proj, dtype="f"),
                "y_translation": 5e-4 * np.ones(self.n_proj, dtype="f"),
                "z_translation": 5e-4 * np.ones(self.n_proj, dtype="f"),
            },
        }
        default_params.update(other_params)
        self.other_params = default_params

    def generate_dict(self):
        beam_group = {
            "incident_energy": self.incident_energy,
        }
        detector_other_params = self.other_params["detector"]
        detector_group = {
            "count_time": detector_other_params["count_time"],
            "data": self.data_volume,
            "distance": detector_other_params["distance"],
            "field_of_view": detector_other_params["field_of_view"],
            "image_key": self.image_key,
            "image_key_control": detector_other_params["image_key_control"],
            "x_pixel_size": detector_other_params["x_pixel_size"],
            "y_pixel_size": detector_other_params["y_pixel_size"],
            "x_magnified_pixel_size": detector_other_params["x_magnified_pixel_size"],
            "y_magnified_pixel_size": detector_other_params["y_magnified_pixel_size"],
        }
        sample_other_params = self.other_params["sample"]
        sample_group = {
            "name": sample_other_params["name"],
            "rotation_angle": self.rotation_angle,
            "x_translation": sample_other_params["x_translation"],
            "y_translation": sample_other_params["y_translation"],
            "z_translation": sample_other_params["z_translation"],
        }
        self.dataset_dict = {
            "beam": beam_group,
            "instrument": {
                "detector": detector_group,
            },
            "sample": sample_group,
        }

    def generate_hdf5_file(self, fname, h5path=None):
        self.fname = fname
        h5path = h5path or "/entry"
        if self.dataset_dict is None:
            self.generate_dict()
        dicttoh5(self.dataset_dict, fname, h5path=h5path, mode="a")
        # Patch the "data" field which is exported as string by dicttoh5 (?!)
        self.dataset_path = os.path.join(h5path, "instrument/detector/data")
        with HDF5File(fname, "a") as fid:
            del fid[self.dataset_path]
            fid[self.dataset_path] = self.dataset_dict["instrument"]["detector"]["data"]

    # Mock some of the HDF5DatasetAnalyzer attributes
    @property
    def dataset_hdf5_url(self):
        if self.fname is None:
            raise ValueError("generate_hdf5_file() was not called")
        return DataUrl(file_path=self.fname, data_path=self.dataset_path, scheme="silx")

    def _get_images_with_key(self, key):
        indices = np.arange(self.image_key.size)[self.image_key == key]
        urls = [
            DataUrl(
                file_path=self.fname,
                data_path=self.dataset_path,
                data_slice=slice(img_idx, img_idx + 1),
                scheme="silx",
            )
            for img_idx in indices
        ]
        return dict(zip(indices, urls))

    @property
    def flats(self):
        return self._get_images_with_key(1)

    @property
    def darks(self):
        return self._get_images_with_key(2)

    def get_data_slices(self, what):
        images = getattr(self, what)
        # we can't directly use set() on slice() object (unhashable). Use tuples
        tuples_list = list(
            set((du.data_slice().start, du.data_slice().stop) for du in get_compacted_dataslices(images).values())
        )
        slices_list = [slice(item[0], item[1]) for item in tuples_list]
        return slices_list


# To be improved
def generate_nx_dataset(out_fname, image_key, data_volume=None, rotation_angle=None):
    nx_template_file = get_file("dummy.nx.tar.gz")
    nx_dict = nxtodict(nx_template_file)
    nx_entry = nx_dict["entry"]

    def _get_field(dict_, path):
        if path.startswith("/"):
            path = path[1:]
        if path.endswith("/"):
            path = path[:-1]
        split_path = path.split("/")
        if len(split_path) == 1:
            return dict_[split_path[0]]
        return _get_field(dict_[split_path[0]], "/".join(split_path[1:]))

    for name in ["image_key", "image_key_control"]:
        nx_entry["data"][name] = image_key
        nx_entry["instrument"]["detector"][name] = image_key

    if rotation_angle is not None:
        nx_entry["data"]["rotation_angle"] = rotation_angle
        nx_entry["sample"]["rotation_angle"] = rotation_angle

    dicttonx(nx_dict, out_fname)



from math import pi
from bisect import bisect
import numpy as np
from ..utils import generate_powers, get_decay, check_supported, get_num_threads, deprecation_warning

#
from .ctf import CTFPhaseRetrieval

#


def lmicron_to_db(Lmicron, energy, distance):
    """
    Utility to convert the "Lmicron" parameter of PyHST
    to a value of delta/beta.

    Parameters
    -----------
    Lmicron: float
        Length in microns, values of the parameter "PAGANIN_Lmicron"
        in PyHST2 parameter file.
    energy: float
        Energy in keV.
    distance: float
        Sample-detector distance in microns

    Notes
    --------
    The conversion is done using the formula

    .. math::
       L^2 = \\pi \\lambda D \\frac{\\delta}{\\beta}

    """
    L2 = Lmicron**2
    wavelength = 1.23984199e-3 / energy
    return L2 / (pi * wavelength * distance)


class PaganinPhaseRetrieval:
    available_padding_modes = ["zeros", "mean", "edge", "symmetric", "reflect"]
    powers = generate_powers()

    def __init__(
        self,
        shape,
        distance=0.5,
        energy=20,
        delta_beta=250.0,
        pixel_size=1e-6,
        padding="edge",
        margin=None,
        use_rfft=True,
        use_R2C=None,
        fftw_num_threads=None,
    ):
        """
        Paganin Phase Retrieval for an infinitely distant point source.
        Formula (10) in [1].

        Parameters
        ----------
        shape: int or tuple
            Shape of each radio, in the format (num_rows, num_columns), i.e
            (size_vertical, size_horizontal).
            If an integer is provided, the shape is assumed to be square.
        distance : float, optional
            Propagation distance in meters.
        energy : float, optional
            Energy in keV.
        delta_beta: float, optional
            delta/beta ratio, where n = (1 - delta) + i*beta is the complex
            refractive index of the sample.
        pixel_size : float, optional
            Detector pixel size in meters. Default is 1e-6 (one micron)
        padding : str, optional
            Padding method. Available are "zeros", "mean", "edge", "sym",
            "reflect". Default is "edge".
            Please refer to the "Padding" section below for more details.
        margin: tuple, optional
            The user may provide integers values U, D, L, R as a tuple under the
            form ((U, D), (L, R)) (same syntax as numpy.pad()).
            The resulting filtered radio will have a size equal to
            (size_vertic - U - D, size_horiz - L - R).
            These values serve to create a "margin" for the filtering process,
            where U, D, L R are the margin of the Up, Down, Left and Right part,
            respectively.
            The filtering is done on a subset of the input radio. The subset
            size is (Nrows - U - D, Ncols - R - L).
            The margins is used to do the padding for the rest of the padded
            array.

            For example in one dimension, where ``padding="edge"``::

               <------------------------------ padded_size --------------------------->
               [padding=edge | padding=data | radio data | padding=data | padding=edge]
               <------ N2 ---><----- L -----><- (N-L-R)--><----- R -----><----- N2 --->

            Some or all the values U, D, L, R can be 0. In this case,
            the padding of the parts related to the zero values will
            fall back to the one of "padding" parameter.
            For example, if padding="edge" and L, R are 0, then
            the left and right parts will be padded with the edges, while
            the Up and Down parts will be padded using the the user-provided
            margins of the radio, and the final data will have shape
            (Nrows - U - D, Ncols).
            Some or all the values U, D, L, R can be the string "auto".
            In this case, the values of U, D, L, R are automatically computed
            as a function of the Paganin filter width.
        use_rfft: bool, optional
            Whether to use Real-to-Complex (R2C) transform instead of
            standard Complex-to-Complex transform, providing better performances
        use_R2C: bool, optional
            DEPRECATED, use use_rfft instead
        fftw_num_threads: bool or None or int, optional
            Whether to use FFTW for speeding up FFT.
            Default is to use all available threads. You can pass a negative number
            to use N - fftw_num_threads cores.

        Important
        ----------
        Mind the units! Distance and pixel size are in meters, and energy is in keV.


        Notes
        ------

        **Padding methods**

        The phase retrieval is a convolution done in Fourier domain using FFT,
        so the Fourier transform size has to be at least twice the size of
        the original data. Mathematically, the data should be padded with zeros
        before being Fourier transformed. However, in practice, this can lead
        to artefacts at the edges (Gibbs effect) if the data does not go to
        zero at the edges.
        Apart from applying an apodization (Hamming, Blackman, etc), a common
        strategy to avoid these artefacts is to pad the data.
        In tomography reconstruction, this is usually done by replicating the
        last(s) value(s) of the edges ; but one can think of other methods:

           - "zeros": the data is simply padded with zeros.
           - "mean": the upper side of extended data is padded with the mean of
             the first row, the lower side with the mean of the last row, etc.
           - "edge": the data is padded by replicating the edges.
             This is the default mode.
           - "sym": the data is padded by mirroring the data with respect
             to its edges. See ``numpy.pad()``.
           - "reflect": the data is padded by reflecting the data with respect
             to its edges, including the edges. See ``numpy.pad()``.


        **Formulas**

        The radio is divided, in the Fourier domain, by the original "Paganin filter" `[1]`.

        .. math::

          F = 1 + \\frac{\\delta}{\\beta} \\lambda D \\pi |k|^2

        where k is the wave vector.


        References
        -----------
        [1] D. Paganin Et Al, "Simultaneous phase and amplitude extraction
            from a single defocused image of a homogeneous object",
            Journal of Microscopy, Vol 206, Part 1, 2002
        """
        self._init_parameters(distance, energy, pixel_size, delta_beta, padding)
        self._calc_shape(shape, margin)
        # COMPAT.
        if use_R2C is not None:
            deprecation_warning("'use_R2C' is replaced with 'use_rfft'", func_name="pag_r2c")
        # -
        self._get_fft(use_rfft, fftw_num_threads)
        self.compute_filter()

    def _init_parameters(self, distance, energy, pixel_size, delta_beta, padding):
        self.distance_cm = distance * 1e2
        self.distance_micron = distance * 1e6
        self.energy_kev = energy
        self.pixel_size_micron = pixel_size * 1e6
        self.delta_beta = delta_beta
        self.wavelength_micron = 1.23984199e-3 / self.energy_kev
        self.padding = padding
        self.padding_methods = {
            "zeros": self._pad_zeros,
            "mean": self._pad_mean,
            "edge": self._pad_edge,
            "symmetric": self._pad_sym,
            "reflect": self._pad_reflect,
        }

    def _get_fft(self, use_rfft, fftw_num_threads):
        self.use_rfft = use_rfft
        self.use_R2C = use_rfft  # Compat.
        fftw_num_threads = get_num_threads(fftw_num_threads)
        if self.use_rfft:
            self.fft_func = np.fft.rfft2
            self.ifft_func = np.fft.irfft2
        else:
            self.fft_func = np.fft.fft2
            self.ifft_func = np.fft.ifft2
        self.use_fftw = False
        if self.use_rfft and (fftw_num_threads > 0):
            # importing silx.math.fft creates opencl contexts all over the place
            # because of the silx.opencl.ocl singleton.
            # So, import silx as late as possible
            from silx.math.fft.fftw import FFTW, __have_fftw__

            if __have_fftw__:
                self.use_fftw = True
                self.fftw = FFTW(shape=self.shape_padded, dtype="f", num_threads=fftw_num_threads)
                self.fft_func = self.fftw.fft
                self.ifft_func = self.fftw.ifft

    def _calc_shape(self, shape, margin):
        if np.isscalar(shape):
            shape = (shape, shape)
        else:
            assert len(shape) == 2
        self.shape = shape
        self._set_margin_value(margin)
        self._calc_padded_shape()

    def _set_margin_value(self, margin):
        self.margin = margin
        if margin is None:
            self.shape_inner = self.shape
            self.use_margin = False
            self.margin = ((0, 0), (0, 0))
            return
        self.use_margin = True
        try:
            ((U, D), (L, R)) = margin
        except ValueError:
            raise ValueError("Expected margin in the format ((U, D), (L, R))")
        for val in [U, D, L, R]:
            if isinstance(val, str) and val != "auto":
                raise ValueError("Expected either an integer, or 'auto'")
            if int(val) != val or val < 0:
                raise ValueError("Expected positive integers for margin values")
        self.shape_inner = (self.shape[0] - U - D, self.shape[1] - L - R)

    def _calc_padded_shape(self):
        """
        Compute the padded shape.
        If margin = 0, length_padded = next_power(2*length).
        Otherwise : length_padded = next_power(2*(length - margins))

        Principle
        ----------

        <--------------------- nx_p --------------------->
        |         |        original data       |         |
        < -- Pl - ><-- L -->< -- nx --><-- R --><-- Pr -->
                   <----------- nx0 ----------->

        Pl, Pr : left/right padding length
        L, R : left/right margin
        nx : length of inner data (and length of final result)
        nx0 : length of original data
        nx_p : total length of padded data
        """
        n_y, n_x = self.shape_inner
        n_y0, n_x0 = self.shape
        n_y_p = self._get_next_power(max(2 * n_y, n_y0))
        n_x_p = self._get_next_power(max(2 * n_x, n_x0))
        self.shape_padded = (n_y_p, n_x_p)
        self.data_padded = np.zeros((n_y_p, n_x_p), dtype=np.float64)

        ((U, D), (L, R)) = self.margin
        n_y0, n_x0 = self.shape
        self.pad_top_len = (n_y_p - n_y0) // 2
        self.pad_bottom_len = n_y_p - n_y0 - self.pad_top_len
        self.pad_left_len = (n_x_p - n_x0) // 2
        self.pad_right_len = n_x_p - n_x0 - self.pad_left_len

    def _get_next_power(self, n):
        """
        Given a number, get the closest (upper) number p such that
        p is a power of 2, 3, 5 and 7.
        """
        idx = bisect(self.powers, n)
        if self.powers[idx - 1] == n:
            return n
        return self.powers[idx]

    def compute_filter(self):
        nyp, nxp = self.shape_padded
        fftfreq = np.fft.rfftfreq if self.use_rfft else np.fft.fftfreq
        fy = np.fft.fftfreq(nyp, d=self.pixel_size_micron)
        fx = fftfreq(nxp, d=self.pixel_size_micron)
        self._coords_grid = np.add.outer(fy**2, fx**2)
        #
        k2 = self._coords_grid
        D = self.distance_micron
        L = self.wavelength_micron
        db = self.delta_beta
        self.paganin_filter = 1.0 / (1 + db * L * D * pi * k2)

    def pad_with_values(self, data, top_val=0, bottom_val=0, left_val=0, right_val=0):
        """
        Pad the data into `self.padded_data` with values.

        Parameters
        ----------
        data: numpy.ndarray
            data (radio)
        top_val: float or numpy.ndarray, optional
            Value(s) to fill the top of the padded data with.
        bottom_val: float or numpy.ndarray, optional
            Value(s) to fill the bottom of the padded data with.
        left_val: float or numpy.ndarray, optional
            Value(s) to fill the left of the padded data with.
        right_val: float or numpy.ndarray, optional
            Value(s) to fill the right of the padded data with.
        """
        self.data_padded.fill(0)
        Pu, Pd = self.pad_top_len, self.pad_bottom_len
        Pl, Pr = self.pad_left_len, self.pad_right_len
        self.data_padded[:Pu, :] = top_val
        self.data_padded[-Pd:, :] = bottom_val
        self.data_padded[:, :Pl] = left_val
        self.data_padded[:, -Pr:] = right_val
        self.data_padded[Pu:-Pd, Pl:-Pr] = data
        # Transform the data to the FFT layout
        self.data_padded = np.roll(self.data_padded, (-Pu, -Pl), axis=(0, 1))

    def _pad_zeros(self, data):
        return self.pad_with_values(data, top_val=0, bottom_val=0, left_val=0, right_val=0)

    def _pad_mean(self, data):
        """
        Pad the data at each border with a different constant value.
        The value depends on the padding size:
          - On the left, value = mean(first data column)
          - On the right, value = mean(last data column)
          - On the top, value = mean(first data row)
          - On the bottom, value = mean(last data row)
        """
        return self.pad_with_values(
            data,
            top_val=np.mean(data[0, :]),
            bottom_val=np.mean(data[-1, :]),
            left_val=np.mean(data[:, 0]),
            right_val=np.mean(data[:, -1]),
        )

    def _pad_numpy(self, data, mode):
        data_padded = np.pad(
            data, ((self.pad_top_len, self.pad_bottom_len), (self.pad_left_len, self.pad_right_len)), mode=mode
        )
        # Transform the data to the FFT layout
        Pu, Pl = self.pad_top_len, self.pad_left_len
        return np.roll(data_padded, (-Pu, -Pl), axis=(0, 1))

    def _pad_edge(self, data):
        self.data_padded = self._pad_numpy(data, mode="edge")

    def _pad_sym(self, data):
        self.data_padded = self._pad_numpy(data, mode="symmetric")

    def _pad_reflect(self, data):
        self.data_padded = self._pad_numpy(data, mode="reflect")

    def pad_data(self, data, padding_method=None):
        padding_method = padding_method or self.padding
        check_supported(padding_method, self.available_padding_modes, "padding mode")
        if padding_method not in self.padding_methods:
            raise ValueError(
                "Unknown padding method %s. Available are: %s"
                % (padding_method, str(list(self.padding_methods.keys())))
            )
        pad_func = self.padding_methods[padding_method]
        pad_func(data)
        return self.data_padded

    def apply_filter(self, radio, padding_method=None, output=None):
        self.pad_data(radio, padding_method=padding_method)
        radio_f = self.fft_func(self.data_padded)
        radio_f *= self.paganin_filter
        radio_filtered = self.ifft_func(radio_f).real
        s0, s1 = self.shape_inner
        ((U, _), (L, _)) = self.margin
        if output is None:
            return radio_filtered[U : U + s0, L : L + s1]
        else:
            output[:, :] = radio_filtered[U : U + s0, L : L + s1]
            return output

    def lmicron_to_db(self, Lmicron):
        """
        Utility to convert the "Lmicron" parameter of PyHST
        to a value of delta/beta.
        Please see the doc of nabu.preproc.phase.lmicron_to_db()
        """
        return lmicron_to_db(Lmicron, self.energy_kev, self.distance_micron)

    __call__ = apply_filter

    retrieve_phase = apply_filter


def compute_paganin_margin(shape, cutoff=1e3, **pag_kwargs):
    """
    Compute the convolution margin to use when calling PaganinPhaseRetrieval class.

    Parameters
    -----------
    shape: tuple
        Detector shape in the form (n_z, n_x)
    """
    P = PaganinPhaseRetrieval(shape, **pag_kwargs)

    ifft_func = np.fft.irfft2 if P.use_rfft else np.fft.ifft2
    conv_kernel = ifft_func(P.paganin_filter)

    vmax = conv_kernel[0, 0]

    v_margin = get_decay(conv_kernel[:, 0], cutoff=cutoff, vmax=vmax)
    h_margin = get_decay(conv_kernel[0, :], cutoff=cutoff, vmax=vmax)
    # If the Paganin filter is very narrow, then the corresponding convolution
    # kernel is constant, and np.argmax() gives 0 (when it should give the max value)
    if v_margin == 0:
        v_margin = shape[0]
    if h_margin == 0:
        h_margin = shape[1]

    return v_margin, h_margin



from typing import Optional, Union

import numpy
from nabu.stitching.overlap import OverlapStichingStrategy, ZStichOverlapKernel
from tomoscan.scanbase import TomoScanBase
import logging
from scipy.ndimage import shift as scipy_shift

_logger = logging.getLogger(__name__)


try:
    from skimage.registration import phase_cross_correlation
except ImportError:
    _logger.warning(
        "Unable to load skimage. Please install it if you want to use it for finding shifts from `find_relative_shifts`"
    )
    __has_sk_phase_correlation__ = False
else:
    __has_sk_phase_correlation__ = True


def test_overlap_stitching_strategy(overlap_1, overlap_2, stitching_strategies):
    """
    stitch the two ovrelap with all the requested strategies.
    Return a dictionary with stitching strategy as key and a result dict as value.
    result dict keys are: 'weights_overlap_1', 'weights_overlap_2', 'stiching'
    """
    res = {}
    for strategy in stitching_strategies:
        s = OverlapStichingStrategy.from_value(strategy)
        stitcher = ZStichOverlapKernel(
            stitching_strategy=s,
            overlap_height=overlap_1.shape[0],
            frame_width=overlap_1.shape[1],
        )
        stiched_overlap, w1, w2 = stitcher.stitch(overlap_1, overlap_2, check_input=True)
        res[s.value] = {
            "stitching": stiched_overlap,
            "weights_overlap_1": w1,
            "weights_overlap_2": w2,
        }
    return res


def find_relative_shifts(
    scan_0: TomoScanBase,
    scan_1: TomoScanBase,
    projection_for_shift: Union[int, str] = "middle",
    invert_order: bool = False,
    x_cross_correlation_function=None,
    y_cross_correlation_function=None,
    auto_flip: bool = True,
) -> tuple:
    """
    deduce the relative shift between the two scans.
    Expected behavior:
    * compute expected overlap area from z_translations and (sample) pixel size
    * call a cross correlation function from the overlap area to compute the x shift and polish the y shift from `projection_for_shift`

    :param TomoScanBase scan_0:
    :param TomoScanBase scan_1:
    :param Union[int,str] projection_for_shift: index fo the projection to use (in projection space or in scan space ?. For now in projection) or str. If str must be in (`middle`, `first`, `last`)
    :param str x_cross_correlation_function: optional method to refine x shift from computing cross correlation. For now valid values are: ("skimage", "skimage-fourier")
    :param str y_cross_correlation_function: optional method to refine y shift from computing cross correlation. For now valid values are: ("skimage", "skimage-fourier")
    :param int minimal_overlap_area_for_cross_correlation: if first approximated overlap shift found from z_translation is lower than this value will fall back on taking the full image for the cross correlation and log a warning
    :param bool invert_order: are projections inverted between the two scans (case if rotation angle are inverted)
    :param bool auto_flip: if True then will automatically flip frames to get a "homogeneous" result based on unflipped frames
    :return: relative shift of scan_1 with scan_0 as reference: (y_shift, x_shift)
    :rtype: tuple

    :warning: this function will flip left-right and up-down frames by default. So it will return shift according to this information
    """

    def get_flat_fielded_proj(scan: TomoScanBase, proj_index: int, reverse: bool, revert_x: bool, revert_y):
        first_proj_idx = sorted(scan_1.projections.keys(), reverse=reverse)[proj_index]
        ff = scan.flat_field_correction(
            (scan.projections[first_proj_idx],),
            (first_proj_idx,),
        )[0]
        if auto_flip and revert_x:
            ff = numpy.fliplr(ff)
        if auto_flip and revert_y:
            ff = numpy.flipud(ff)
        return ff

    if isinstance(projection_for_shift, str):
        if projection_for_shift.lower() == "first":
            projection_for_shift = 0
        elif projection_for_shift.lower() == "last":
            projection_for_shift = -1
        elif projection_for_shift.lower() == "middle":
            projection_for_shift = len(scan_0.projections) // 2
        else:
            try:
                projection_for_shift = int(projection_for_shift)
            except ValueError:
                raise ValueError(
                    f"{projection_for_shift} cannot be cast to an int and is not one of the possible ('first', 'last', 'middle')"
                )
    elif not isinstance(projection_for_shift, (int, numpy.number)):
        raise TypeError(
            f"projection_for_shift is expected to be an int. Not {type(projection_for_shift)} - {projection_for_shift}"
        )

    proj_0 = get_flat_fielded_proj(
        scan_0,
        projection_for_shift,
        reverse=False,
        revert_x=scan_0.get_x_flipped(default=False),
        revert_y=scan_0.get_y_flipped(default=False),
    )
    proj_1 = get_flat_fielded_proj(
        scan_1,
        projection_for_shift,
        reverse=invert_order,
        revert_x=scan_1.get_x_flipped(default=False),
        revert_y=scan_1.get_y_flipped(default=False),
    )

    # get overlap area from z
    scan_0_y_bb = scan_0.get_bounding_box(axis="z")
    scan_1_y_bb = scan_1.get_bounding_box(axis="z")

    scan_0_scan_1_overlap = scan_0_y_bb.get_overlap(scan_1_y_bb)

    if scan_0_scan_1_overlap is not None:
        overlap = scan_0_scan_1_overlap.max - scan_0_scan_1_overlap.min
        overlap_percentage = (overlap) / (scan_0_y_bb.max - scan_0_y_bb.min)
        y_overlap_frm_position_in_pixel = int(overlap_percentage * scan_0.dim_2)
        overlap_1 = proj_0[-y_overlap_frm_position_in_pixel:]
        overlap_2 = proj_1[:y_overlap_frm_position_in_pixel:]
    else:
        _logger.warning(
            "no overlap founds from scan metadata. Take the full image to try to find an overlap. Automatic shift deduction has an higher probability to fail"
        )

    x_found_shift = 0
    y_found_shift = -y_overlap_frm_position_in_pixel

    if x_cross_correlation_function in ("skimage", "skimage-fourier"):
        if not __has_sk_phase_correlation__:
            raise ValueError("scikit-image not installed. Cannot do phase correlation from it")
        if x_cross_correlation_function == "skimage-fourier":
            overlap_1 = numpy.fft.fftn(overlap_1)
            overlap_2 = numpy.fft.fftn(overlap_2)
            space = "fourier"
        else:
            space = "real"
        found_shift, _, _ = phase_cross_correlation(reference_image=overlap_1, moving_image=overlap_2, space=space)
        x_found_shift = found_shift[1]
    elif x_cross_correlation_function is not None:
        raise ValueError(f"requested cross correlation function not handled ({x_cross_correlation_function})")

    if y_cross_correlation_function in ("skimage", "skimage-fourier"):
        if not __has_sk_phase_correlation__:
            raise ValueError("scikit-image not installed. Cannot do phase correlation from it")
        if y_cross_correlation_function == "skimage-fourier":
            overlap_1 = numpy.fft.fftn(overlap_1)
            overlap_2 = numpy.fft.fftn(overlap_2)
            space = "fourier"
        else:
            space = "real"
        found_shift, _, _ = phase_cross_correlation(reference_image=overlap_1, moving_image=overlap_2, space=space)
        y_found_shift = found_shift[0] - y_overlap_frm_position_in_pixel
    elif y_cross_correlation_function is not None:
        raise ValueError(f"requested cross correlation function not handled ({y_cross_correlation_function})")

    #
    if y_found_shift > 0:
        _logger.warning(
            f"found a positive shift ({found_shift[0]}) when a negative one is expected. Are you sure about the scan z ordering. This is likely z stitching will fails"
        )

    return tuple([int(y_found_shift), int(x_found_shift)])



from os import path
from tempfile import TemporaryDirectory
import pytest
import numpy as np
from tifffile import TiffReader
from silx.io.dictdump import h5todict, dicttoh5
from nabu.misc.utils import psnr
from nabu.io.writer import NXProcessWriter, TIFFWriter, JP2Writer, __have_jp2k__
from nabu.io.reader import import_h5_to_dict
from nabu.testutils import get_data

if __have_jp2k__:
    from glymur import Jp2k


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls._tmpdir = TemporaryDirectory(prefix="nabu_")
    cls.tempdir = cls._tmpdir.name
    cls.sino_data = get_data("mri_sino500.npz")["data"].astype(np.uint16)
    cls.data = cls.sino_data
    yield
    # teardown
    cls._tmpdir.cleanup()


@pytest.mark.usefixtures("bootstrap")
class TestTiff:
    def _check_tif_file(self, fname, expected_data, n_expected_images):
        with TiffReader(fname) as tif:
            assert len(tif.pages) == n_expected_images
            for i in range(n_expected_images):
                data_read = tif.pages[i].asarray()
                if expected_data.ndim == 3:
                    expected_data_ = expected_data[i]
                else:
                    expected_data_ = expected_data
                assert np.allclose(data_read, expected_data_)

    def test_2D(self):
        # TODO use start_index=None (by default) in TIFFWriter
        # if start_index is None:
        # url = DataUrl(file_path=dirname(self.fname), data_path="basename(self.fname).tiff", scheme="tifffile")
        # volume = TIFFVolume(volume_basename=basename(self.fname), data_url=url)
        # volume.data = img
        # volume.save()
        pytest.skip("Writing a single 2D tiff is disabled for now")

        fname = path.join(self.tempdir, "test_tiff2D.tif")
        data = np.arange(100 * 101, dtype="f").reshape((100, 101))
        nabu_tif = TIFFWriter(fname)
        nabu_tif.write(data)
        self._check_tif_file(fname, data, 1)

    def test_3D_data_split_in_multiple_files(self):
        fname = path.join(self.tempdir, "test_tiff3D_single.tif")
        data = np.arange(11 * 100 * 101, dtype="f").reshape((11, 100, 101))
        nabu_tif = TIFFWriter(fname, multiframe=False, start_index=500)
        nabu_tif.write(data)

        assert not (path.isfile(fname)), "found %s" % fname

        prefix, ext = path.splitext(fname)
        for i in range(data.shape[0]):
            curr_rel_fname = prefix + str("_%06d" % (i + nabu_tif.start_index)) + ext
            curr_fname = path.join(self.tempdir, curr_rel_fname)
            self._check_tif_file(curr_fname, data[i], 1)

    def test_3D_data_in_one_file(self):
        fname = path.join(self.tempdir, "test_tiff3D_multi.tif")
        data = np.arange(11 * 100 * 101, dtype="f").reshape((11, 100, 101))
        nabu_tif = TIFFWriter(fname, multiframe=True)
        nabu_tif.write(data)

        assert path.isfile(fname)
        self._check_tif_file(fname, data, data.shape[0])


@pytest.mark.skipif(not (__have_jp2k__), reason="Need openjpeg2000/glymur for this test")
@pytest.mark.usefixtures("bootstrap")
class TestJP2:
    def _check_jp2_file(self, fname, expected_data, expected_psnr=None):
        data = Jp2k(fname)[:]
        if expected_psnr is None:
            assert np.allclose(data, expected_data)
        else:
            computed_psnr = psnr(data, expected_data)
            assert np.abs(computed_psnr - expected_psnr) < 1

    def test_2D_lossless(self):
        data = get_data("mri_sino500.npz")["data"].astype(np.uint16)
        fname = path.join(self.tempdir, "sino500.jp2")
        nabu_jp2 = JP2Writer(fname, psnr=[0])
        nabu_jp2.write(data)
        self._check_jp2_file(fname, data)

    def test_2D_lossy(self):
        fname = path.join(self.tempdir, "sino500_lossy.jp2")
        nabu_jp2 = JP2Writer(fname, psnr=[80])
        nabu_jp2.write(self.sino_data)
        self._check_jp2_file(fname, self.sino_data, expected_psnr=80)

    def test_3D(self):
        fname = path.join(self.tempdir, "sino500_multi.jp2")
        n_images = 5
        data = np.tile(self.sino_data, (n_images, 1, 1))
        for i in range(n_images):
            data[i] += i

        nabu_jp2 = JP2Writer(fname, start_index=10)
        nabu_jp2.write(data)

        assert not (path.isfile(fname))

        prefix, ext = path.splitext(fname)
        for i in range(data.shape[0]):
            curr_rel_fname = prefix + str("_%06d" % (i + nabu_jp2.start_index)) + ext
            curr_fname = path.join(self.tempdir, curr_rel_fname)
            self._check_jp2_file(curr_fname, data[i])


@pytest.fixture(scope="class")
def bootstrap_h5(request):
    cls = request.cls
    cls._tmpdir = TemporaryDirectory(prefix="nabu_")
    cls.tempdir = cls._tmpdir.name
    cls.data = get_data("mri_sino500.npz")["data"].astype(np.uint16)
    cls.h5_config = {
        "key1": "value1",
        "some_int": 1,
        "some_float": 1.0,
        "some_dict": {
            "numpy_array": np.ones((5, 6), dtype="f"),
            "key2": "value2",
        },
    }
    yield
    # teardown
    cls._tmpdir.cleanup()


@pytest.mark.usefixtures("bootstrap_h5")
class TestNXWriter:
    def test_write_simple(self):
        fname = path.join(self.tempdir, "sino500.h5")
        writer = NXProcessWriter(fname, entry="entry0000")
        writer.write(self.data, "test_write_simple")

    def test_write_with_config(self):
        fname = path.join(self.tempdir, "sino500_cfg.h5")
        writer = NXProcessWriter(fname, entry="entry0000")
        writer.write(self.data, "test_write_with_config", config=self.h5_config)

    def test_overwrite(self):
        fname = path.join(self.tempdir, "sino500_overwrite.h5")
        writer = NXProcessWriter(fname, entry="entry0000", overwrite=True)
        writer.write(self.data, "test_overwrite", config=self.h5_config)

        writer2 = NXProcessWriter(fname, entry="entry0001", overwrite=True)
        writer2.write(self.data, "test_overwrite", config=self.h5_config)

        # overwrite entry0000
        writer3 = NXProcessWriter(fname, entry="entry0000", overwrite=True)
        new_data = self.data.copy()
        new_data += 1
        new_config = self.h5_config.copy()
        new_config["key1"] = "modified value"
        writer3.write(new_data, "test_overwrite", config=new_config)

        res = import_h5_to_dict(fname, "/")
        assert "entry0000" in res
        assert "entry0001" in res
        assert np.allclose(res["entry0000"]["test_overwrite"]["results"]["data"], self.data + 1)
        rec_cfg = res["entry0000"]["test_overwrite"]["configuration"]
        assert rec_cfg["key1"] == "modified value"

    def test_no_overwrite(self):
        fname = path.join(self.tempdir, "sino500_no_overwrite.h5")
        writer = NXProcessWriter(fname, entry="entry0000", overwrite=False)
        writer.write(self.data, "test_no_overwrite")

        writer2 = NXProcessWriter(fname, entry="entry0000", overwrite=False)
        with pytest.raises((RuntimeError, OSError)) as ex:
            writer2.write(self.data, "test_no_overwrite")

        message = "Error should have been raised for trying to overwrite, but got the following: %s" % str(ex.value)



import os.path as path
import numpy as np
import pytest
from nabu.testutils import get_data
from nabu.cuda.utils import __has_pycuda__
from nabu.reconstruction.sinogram import SinoNormalization

if __has_pycuda__:
    from nabu.reconstruction.sinogram_cuda import CudaSinoNormalization
    import pycuda.gpuarray as garray


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("sino_refill.npy")
    cls.tol = 1e-7
    cls.norm_array_1D = np.arange(cls.sino.shape[-1]) + 1
    cls.norm_array_2D = np.arange(cls.sino.size).reshape(cls.sino.shape) + 1


@pytest.mark.usefixtures("bootstrap")
class TestSinoNormalization:
    def test_sino_normalization(self):
        sino_proc = SinoNormalization(kind="chebyshev", sinos_shape=self.sino.shape)
        sino = self.sino.copy()
        sino_proc.normalize(sino)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for sinogram normalization with cuda backend")
    def test_sino_normalization_cuda(self):
        sino_proc = SinoNormalization(kind="chebyshev", sinos_shape=self.sino.shape)
        sino = self.sino.copy()
        ref = sino_proc.normalize(sino)

        cuda_sino_proc = CudaSinoNormalization(kind="chebyshev", sinos_shape=self.sino.shape)
        d_sino = garray.to_gpu(self.sino)
        cuda_sino_proc.normalize(d_sino)
        res = d_sino.get()

        assert np.max(np.abs(res - ref)) < self.tol

    def get_normalization_reference_result(self, op, normalization_arr):
        # Perform explicit operations to compare with numpy.divide, numpy.subtract, etc
        if op == "subtraction":
            ref = self.sino - normalization_arr
        elif op == "division":
            ref = self.sino / normalization_arr
        return ref

    def test_sino_array_subtraction_and_division(self):
        with pytest.raises(ValueError):
            SinoNormalization(kind="subtraction", sinos_shape=self.sino.shape)

        def compare_normalizations(normalization_arr, op):
            sino_normalization = SinoNormalization(
                kind=op, sinos_shape=self.sino.shape, normalization_array=normalization_arr
            )
            sino = self.sino.copy()
            sino_normalization.normalize(sino)
            ref = self.get_normalization_reference_result(op, normalization_arr)
            assert np.allclose(sino, ref), "operation=%s, normalization_array dims=%d" % (op, normalization_arr.ndim)

        compare_normalizations(self.norm_array_1D, "subtraction")
        compare_normalizations(self.norm_array_1D, "division")
        compare_normalizations(self.norm_array_2D, "subtraction")
        compare_normalizations(self.norm_array_2D, "division")

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for sinogram normalization with cuda backend")
    def test_sino_array_subtraction_cuda(self):
        with pytest.raises(ValueError):
            CudaSinoNormalization(kind="subtraction", sinos_shape=self.sino.shape)

        def compare_normalizations(normalization_arr, op):
            sino_normalization = CudaSinoNormalization(
                kind=op, sinos_shape=self.sino.shape, normalization_array=normalization_arr
            )
            sino = garray.to_gpu(self.sino)
            sino_normalization.normalize(sino)
            ref = self.get_normalization_reference_result(op, normalization_arr)
            assert np.allclose(sino.get(), ref)

        compare_normalizations(self.norm_array_1D, "subtraction")
        compare_normalizations(self.norm_array_2D, "subtraction")
        compare_normalizations(self.norm_array_1D, "division")
        compare_normalizations(self.norm_array_2D, "division")



from ..utils import updiv
import pycuda.gpuarray as garray
from pycuda.compiler import SourceModule


class CudaKernel:
    """
    Helper class that wraps CUDA kernel through pycuda SourceModule.

    Parameters
    -----------
    kernel_name: str
        Name of the CUDA kernel.
    filename: str, optional
        Path to the file name containing kernels definitions
    src: str, optional
        Source code of kernels definitions
    signature: str, optional
        Signature of kernel function. If provided, pycuda will not guess the types
        of kernel arguments, making the calls slightly faster.
        For example, a function acting on two pointers, an integer and a float32
        has the signature "PPif".
    texrefs: list, optional
        List of texture references, if any
    automation_params: dict, optional
        Automation parameters, see below
    sourcemodule_kwargs: optional
        Extra arguments to provide to pycuda.compiler.SourceModule(),

    Automation parameters
    ----------------------
    automation_params is a dictionary with the following keys and default values.
        guess_block: bool (True)
            If block is not specified during calls, choose a block size based on
            the size/dimensions of the first array.
            Mind that it is unlikely to be the optimal choice.
        guess_grid: bool (True):
            If the grid size is not specified during calls, choose a grid size
            based on the size of the first array.
        follow_gpuarr_ptr: bool (True)
            specify gpuarray.gpudata for all GPUArrays. Otherwise, raise an error.
    """

    def __init__(
        self,
        kernel_name,
        filename=None,
        src=None,
        signature=None,
        texrefs=[],
        automation_params=None,
        **sourcemodule_kwargs,
    ):
        self.check_filename_src(filename, src)
        self.set_automation_params(automation_params)
        self.compile_kernel_source(kernel_name, sourcemodule_kwargs)
        self.prepare(signature, texrefs)

    def check_filename_src(self, filename, src):
        err_msg = "Please provide either filename or src"
        if filename is None and src is None:
            raise ValueError(err_msg)
        if filename is not None and src is not None:
            raise ValueError(err_msg)
        if filename is not None:
            with open(filename) as fid:
                src = fid.read()
        self.filename = filename
        self.src = src

    def set_automation_params(self, automation_params):
        self.automation_params = {
            "guess_block": True,
            "guess_grid": True,
            "follow_gpuarr_ptr": True,
        }
        automation_params = automation_params or {}
        self.automation_params.update(automation_params)

    def compile_kernel_source(self, kernel_name, sourcemodule_kwargs):
        self.sourcemodule_kwargs = sourcemodule_kwargs
        self.kernel_name = kernel_name
        self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
        self.func = self.module.get_function(kernel_name)

    def prepare(self, kernel_signature, texrefs):
        self.prepared = False
        self.kernel_signature = kernel_signature
        self.texrefs = texrefs
        if kernel_signature is not None:
            self.func.prepare(self.kernel_signature, texrefs=texrefs)
            self.prepared = True

    @staticmethod
    def guess_grid_size(shape, block_size):
        # python: (z, y, x) -> cuda: (x, y, z)
        res = tuple(map(lambda x: updiv(x[0], x[1]), zip(shape[::-1], block_size)))
        if len(res) == 2:
            res += (1,)
        return res

    @staticmethod
    def guess_block_size(shape):
        """
        Guess a block size based on the shape of an array.
        """
        ndim = len(shape)
        if ndim == 1:
            return (128, 1, 1)
        if ndim == 2:
            return (32, 32, 1)
        else:
            return (16, 8, 8)

    def get_block_grid(self, *args, **kwargs):
        block = None
        grid = None
        if ("block" not in kwargs) or (kwargs["block"] is None):
            if self.automation_params["guess_block"]:
                block = self.guess_block_size(args[0].shape)
            else:
                raise ValueError("Please provide block size")
        else:
            block = kwargs["block"]
        if ("grid" not in kwargs) or (kwargs["grid"] is None):
            if self.automation_params["guess_grid"]:
                grid = self.guess_grid_size(args[0].shape, block)
            else:
                raise ValueError("Please provide block grid")
        else:
            grid = kwargs["grid"]
        self.last_block_size = block
        self.last_grid_size = grid
        return block, grid

    def follow_gpu_arr(self, args):
        args = list(args)
        # Replace GPUArray with GPUArray.gpudata
        for i, arg in enumerate(args):
            if isinstance(arg, garray.GPUArray):
                args[i] = arg.gpudata
        return tuple(args)

    def get_last_kernel_time(self):
        """
        Return the execution time (in seconds) of the last called kernel.
        The last called kernel should have been called with time_kernel=True.
        """
        if self.last_kernel_time is not None:
            return self.last_kernel_time()
        else:
            return None

    def call(self, *args, **kwargs):
        block, grid = self.get_block_grid(*args, **kwargs)
        # pycuda crashes when any element of block/grid is not a python int (ex. numpy.int64).
        # A weird behavior once observed is "data.shape" returning (np.int64, int, int) (!).
        # Ensure that everything is a python integer.
        block = tuple(int(x) for x in block)
        grid = tuple(int(x) for x in grid)
        #
        args = self.follow_gpu_arr(args)

        if self.prepared:
            func_call = self.func.prepared_call
            if "time_kernel" in kwargs:
                func_call = self.func.prepared_timed_call
                kwargs.pop("time_kernel")
            if "block" in kwargs:
                kwargs.pop("block")
            if "grid" in kwargs:
                kwargs.pop("grid")
            t = func_call(grid, block, *args, **kwargs)
        else:
            kwargs["block"] = block
            kwargs["grid"] = grid
            t = self.func(*args, **kwargs)
        # ~ return t
        # TODO return event like in OpenCL ?
        self.last_kernel_time = t  # list ?

    __call__ = call



from .. import version
from ..resources.utils import is_hdf5_extension
from ..pipeline.config import parse_nabu_config_file
from ..pipeline.config_validators import convert_to_int
from .cli_configs import ReconstructConfig
from .utils import parse_params_values
from .reconstruct import update_reconstruction_start_end, get_log_file


def main_helical():
    ReconstructConfig["dry_run"] = {
        "help": "Stops after printing some information on the reconstruction layout.",
        "default": 0,
        "type": int,
    }
    args = parse_params_values(
        ReconstructConfig,
        parser_description=f"Perform a helical tomographic reconstruction",
        program_version="nabu " + version,
    )

    # Imports are done here, otherwise "nabu --version" takes forever
    from ..pipeline.helical.processconfig import ProcessConfig
    from ..pipeline.helical.helical_reconstruction import HelicalReconstructorRegridded

    #

    # A crash with scikit-cuda happens only on PPC64 platform if and nvidia-persistenced is running.
    # On such machines, a warm-up has to be done.
    import platform

    if platform.machine() == "ppc64le":
        try:
            from silx.math.fft.cufft import CUFFT
        except:  # can't catch narrower - cublasNotInitialized requires cublas !
            CUFFT = None
    #

    logfile = get_log_file(args["logfile"], args["log_file"], forbidden=[args["input_file"]])
    conf_dict = parse_nabu_config_file(args["input_file"])
    update_reconstruction_start_end(conf_dict, args["slice"].strip())

    proc = ProcessConfig(conf_dict=conf_dict, create_logger=logfile)
    logger = proc.logger

    if "rotate_projections" in proc.processing_steps:
        message = """ The rotate_projections step is activated. The Helical pipelines are not yet suited for  projection rotation
        it will soon be implemented. For the moment
        you should desactivate the rotation options in nabu.conf
        """
        raise ValueError(message)

    # Determine which reconstructor to use
    reconstructor_cls = None
    phase_method = None
    if "phase" in proc.processing_steps:
        phase_method = proc.processing_options["phase"]["method"]

    # fix the reconstruction roi if not given
    if "reconstruction" in proc.processing_steps:
        rec_config = proc.processing_options["reconstruction"]

        rot_center = rec_config["rotation_axis_position"]
        Nx, Ny = proc.dataset_info.radio_dims

        if proc.nabu_config["reconstruction"]["auto_size"]:
            if 2 * rot_center > Nx:
                w = int(round(2 * rot_center))
            else:
                w = int(round(2 * Nx - 2 * rot_center))
            rec_config["start_x"] = int(round(rot_center - w / 2))
            rec_config["end_x"] = int(round(rot_center + w / 2))

            rec_config["start_y"] = rec_config["start_x"]
            rec_config["end_y"] = rec_config["end_x"]

    reconstructor_cls = HelicalReconstructorRegridded

    logger.debug("Using pipeline: %s" % reconstructor_cls.__name__)

    # Get extra options
    extra_options = {
        "gpu_mem_fraction": args["gpu_mem_fraction"],
        "cpu_mem_fraction": args["cpu_mem_fraction"],
    }
    extra_options.update(
        {
            ##### ??? "use_phase_margin": args["use_phase_margin"],
            "max_chunk_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
            "phase_margin": args["phase_margin"],
            "dry_run": args["dry_run"],
        }
    )

    R = reconstructor_cls(proc, logger=logger, extra_options=extra_options)

    R.reconstruct()
    if not R.dry_run:
        R.merge_data_dumps()
        if is_hdf5_extension(proc.nabu_config["output"]["file_format"]):
            R.merge_hdf5_reconstructions()
        R.merge_histograms()


if __name__ == "__main__":
    main_helical()



from glob import glob
from os import path, getcwd, chdir
from posixpath import join as posix_join
from datetime import datetime
import numpy as np
from h5py import VirtualSource, VirtualLayout
from silx.io.dictdump import dicttoh5, dicttonx
from silx.io.url import DataUrl

try:
    from tomoscan.io import HDF5File
except:
    from h5py import File as HDF5File
from tomoscan.esrf import EDFVolume, HDF5Volume, TIFFVolume, MultiTIFFVolume, JP2KVolume, RawVolume
from tomoscan.esrf.volume.jp2kvolume import has_glymur as __have_jp2k__
from .. import version as nabu_version
from ..utils import merged_shape, deprecation_warning
from ..misc.utils import rescale_data
from .utils import convert_dict_values


def get_datetime():
    """
    Function used by some writers to indicate the current date.
    """
    return datetime.now().replace(microsecond=0).isoformat()


class Writer:
    """
    Base class for all writers.
    """

    def __init__(self, fname):
        self.fname = fname

    def get_filename(self):
        return self.fname


class TomoscanNXProcessWriter(Writer):
    """
    A class to write Nexus file with a processing result - using tomoscan.volumes as a backend
    """

    def __init__(self, fname, entry=None, filemode="a", overwrite=False):
        """
        Initialize a NXProcessWriter.

        Parameters
        -----------
        fname: str
            Path to the HDF5 file.
        entry: str, optional
            Entry in the HDF5 file. Default is "entry"
        """
        super().__init__(fname)
        self._set_entry(entry)
        self._filemode = filemode
        # TODO: notify file mode is deprecated
        self.overwrite = overwrite

    def _set_entry(self, entry):
        self.entry = entry or "entry"
        data_path = posix_join("/", self.entry)
        self.data_path = data_path

    def _write_npadday(self, result, volume, nx_info):
        if result.ndim == 2:
            result = result.reshape(1, result.shape[0], result.shape[1])
        volume.data = result

        self._update_volume_metadata(volume)

        volume.save()
        results_path = posix_join(nx_info["nx_process_path"], "results", nx_info["data_name"])
        process_name = nx_info["process_name"]
        process_info = nx_info["process_info"]
        process_info.update(
            {
                f"{process_name}/results@NX_class": "NXdata",
                f"{process_name}/results@signal": nx_info["data_name"],
            }
        )
        if nx_info.get("is_frames_stack", True):
            process_info.update({f"{process_name}/results@interpretation": "image"})
        if nx_info.get("direct_access", False):
            # prepare the direct access plots
            process_info.update(
                {
                    f"{process_name}@default": "results",
                    "@default": f"{process_name}/results",
                }
            )
        return results_path

    def _write_dict(self, result, volume, nx_info):
        self._update_volume_metadata(volume)
        volume.save_metadata()  # if result is a dictionary then we only have some metadata to be saved
        results_path = posix_join(nx_info["nx_process_path"], "results")
        proc_result_key = posix_join(nx_info["process_name"], "results")
        proc_result = convert_dict_values(result, {None: "None"})
        process_info = nx_info["process_info"]
        process_info.update({proc_result_key: proc_result})
        return results_path

    def _write_virtual_layout(self, result, volume, nx_info):
        # TODO: add test on tomoscan to ensure this use case is handled
        volume.data = result
        self._update_volume_metadata(volume)
        volume.save()
        results_path = posix_join(nx_info["nx_process_path"], "results", nx_info["data_name"])
        return results_path

    @staticmethod
    def _update_volume_metadata(volume):
        if volume.metadata is not None:
            volume.metadata = convert_dict_values(
                volume.metadata,
                {None: "None"},
            )

    def write(
        self,
        result,
        process_name,
        processing_index=0,
        config=None,
        data_name="data",
        is_frames_stack=True,
        direct_access=True,
    ) -> str:
        """
        Write the result in the current NXProcess group.

        Parameters
        ----------
        result: numpy.ndarray
            Array containing the processing result
        process_name: str
            Name of the processing
        processing_index: int
            Index of the processing (in a pipeline)
        config: dict, optional
            Dictionary containing the configuration.
        """
        entry_path = self.data_path
        nx_process_path = "/".join([entry_path, process_name])

        if config is not None:
            config.update({"@NX_class": "NXcollection"})

        nabu_process_info = {
            "@NX_class": "NXentry",
            f"{process_name}@NX_class": "NXprocess",
            f"{process_name}/program": "nabu",
            f"{process_name}/version": nabu_version,
            f"{process_name}/date": get_datetime(),
            f"{process_name}/sequence_index": np.int32(processing_index),
        }

        # Create HDF5Volume object with initial information
        volume = HDF5Volume(
            data_url=DataUrl(
                file_path=self.fname,
                data_path=f"{nx_process_path}/results/{data_name}",
                scheme="silx",
            ),
            metadata_url=DataUrl(
                file_path=self.fname,
                data_path=f"{nx_process_path}/configuration",
            ),
            metadata=config,
            overwrite=self.overwrite,
        )
        if isinstance(result, dict):
            write_method = self._write_dict
        elif isinstance(result, np.ndarray):
            write_method = self._write_npadday
        elif isinstance(result, VirtualLayout):
            write_method = self._write_virtual_layout
        else:
            raise TypeError(f"'result' must be a dict, numpy array or h5py.VirtualLayout, not {type(result)}")
        nx_info = {
            "process_name": process_name,
            "nx_process_path": nx_process_path,
            "process_info": nabu_process_info,
            "data_name": data_name,
            "is_frames_stack": is_frames_stack,
            "direct_access": direct_access,
        }
        results_path = write_method(result, volume, nx_info)

        dicttonx(
            nabu_process_info,
            h5file=self.fname,
            h5path=entry_path,
            update_mode="replace",
            mode="a",
        )
        return results_path


###################################################################################################
## Nabu original code for NXProcessWriter - also works for non-3D data, does not depend on tomoscan
###################################################################################################


def h5_write_object(h5group, key, value, overwrite=False, default_val=None):
    existing_val = h5group.get(key, default_val)
    if existing_val is not default_val:
        if not overwrite:
            raise OSError("Unable to create link (name already exists): %s" % h5group.name)
        else:
            h5group.pop(key)
    h5group[key] = value


class NXProcessWriter(Writer):
    """
    A class to write Nexus file with a processing result.
    """

    def __init__(self, fname, entry=None, filemode="a", overwrite=False):
        """
        Initialize a NXProcessWriter.

        Parameters
        -----------
        fname: str
            Path to the HDF5 file.
        entry: str, optional
            Entry in the HDF5 file. Default is "entry"
        """
        super().__init__(fname)
        self._set_entry(entry)
        self._filemode = filemode
        self.overwrite = overwrite

    def _set_entry(self, entry):
        self.entry = entry or "entry"
        data_path = posix_join("/", self.entry)
        self.data_path = data_path

    def write(
        self,
        result,
        process_name,
        processing_index=0,
        config=None,
        data_name="data",
        is_frames_stack=True,
        direct_access=True,
    ):
        """
        Write the result in the current NXProcess group.

        Parameters
        ----------
        result: numpy.ndarray
            Array containing the processing result
        process_name: str
            Name of the processing
        processing_index: int
            Index of the processing (in a pipeline)
        config: dict, optional
            Dictionary containing the configuration.
        """
        swmr = self._filemode == "r"
        with HDF5File(self.fname, self._filemode, swmr=swmr) as fid:
            nx_entry = fid.require_group(self.data_path)
            if "NX_class" not in nx_entry.attrs:
                nx_entry.attrs["NX_class"] = "NXentry"

            nx_process = nx_entry.require_group(process_name)
            nx_process.attrs["NX_class"] = "NXprocess"

            metadata = {
                "program": "nabu",
                "version": nabu_version,
                "date": get_datetime(),
                "sequence_index": np.int32(processing_index),
            }
            for key, val in metadata.items():
                h5_write_object(nx_process, key, val, overwrite=self.overwrite)

            if config is not None:
                export_dict_to_h5(
                    config, self.fname, posix_join(nx_process.name, "configuration"), overwrite_data=True, mode="a"
                )
                nx_process["configuration"].attrs["NX_class"] = "NXcollection"
            if isinstance(result, dict):
                results_path = posix_join(nx_process.name, "results")
                export_dict_to_h5(result, self.fname, results_path, overwrite_data=self.overwrite, mode="a")
            else:
                nx_data = nx_process.require_group("results")
                results_path = nx_data.name
                nx_data.attrs["NX_class"] = "NXdata"
                nx_data.attrs["signal"] = data_name

                results_data_path = posix_join(results_path, data_name)
                if self.overwrite and results_data_path in fid:
                    del fid[results_data_path]

                if isinstance(result, VirtualLayout):
                    nx_data.create_virtual_dataset(data_name, result)
                else:  # assuming array-like
                    nx_data[data_name] = result
                if is_frames_stack:
                    nx_data[data_name].attrs["interpretation"] = "image"
                nx_data.attrs["signal"] = data_name

            # prepare the direct access plots
            if direct_access:
                nx_process.attrs["default"] = "results"
                if "default" not in nx_entry.attrs:
                    nx_entry.attrs["default"] = posix_join(nx_process.name, "results")

            # Return the internal path to "results"
            return results_path


# COMPAT.
LegacyNXProcessWriter = NXProcessWriter
#

########################################################################################
########################################################################################
########################################################################################


def export_dict_to_h5(dic, h5file, h5path, overwrite_data=True, mode="a"):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing None with "None"

    Parameters
    -----------
    dic: dict
        Dictionary containing the options
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    overwrite_data: bool, optional
        Whether to overwrite data when writing HDF5. Default is True
    mode: str, optional
        File mode. Default is "a" (append).
    """
    modified_dic = convert_dict_values(
        dic,
        {None: "None"},
    )
    update_mode = {True: "modify", False: "add"}[bool(overwrite_data)]
    return dicttoh5(modified_dic, h5file=h5file, h5path=h5path, update_mode=update_mode, mode=mode)


def create_virtual_layout(files_or_pattern, h5_path, base_dir=None, axis=0):
    """
    Create a HDF5 virtual layout.

    Parameters
    ----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    """
    prev_cwd = None
    if base_dir is not None:
        prev_cwd = getcwd()
        chdir(base_dir)
    if isinstance(files_or_pattern, str):
        files_list = glob(files_or_pattern)
        files_list.sort()
    else:  # list
        files_list = files_or_pattern
    if files_list == []:
        raise ValueError("Nothing found as pattern %s" % files_or_pattern)
    virtual_sources = []
    shapes = []
    for fname in files_list:
        with HDF5File(fname, "r", swmr=True) as fid:
            shape = fid[h5_path].shape
        vsource = VirtualSource(fname, name=h5_path, shape=shape)
        virtual_sources.append(vsource)
        shapes.append(shape)
    total_shape = merged_shape(shapes, axis=axis)

    virtual_layout = VirtualLayout(shape=total_shape, dtype="f")
    start_idx = 0
    for vsource, shape in zip(virtual_sources, shapes):
        n_imgs = shape[axis]
        # Perhaps there is more elegant
        if axis == 0:
            virtual_layout[start_idx : start_idx + n_imgs] = vsource
        elif axis == 1:
            virtual_layout[:, start_idx : start_idx + n_imgs, :] = vsource
        elif axis == 2:
            virtual_layout[:, :, start_idx : start_idx + n_imgs] = vsource
        else:
            raise ValueError("Only axis 0,1,2 are supported")
        #
        start_idx += n_imgs

    if base_dir is not None:
        chdir(prev_cwd)
    return virtual_layout


def merge_hdf5_files(
    files_or_pattern,
    h5_path,
    output_file,
    process_name,
    output_entry=None,
    output_filemode="a",
    data_name="data",
    processing_index=0,
    config=None,
    base_dir=None,
    axis=0,
    overwrite=False,
):
    """
    Parameters
    -----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    output_file: str
        Path of the output file
    process_name: str
        Name of the process
    output_entry: str, optional
        Output HDF5 root entry (default is "/entry")
    output_filemode: str, optional
        File mode for output file. Default is "a" (append)
    processing_index: int, optional
        Processing index for the output file. Default is 0.
    config: dict, optional
        Dictionary describing the configuration needed to get the results.
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    overwrite: bool, optional
        Whether to overwrite already existing data in the final file.
        Default is False.
    """
    if base_dir is not None:
        prev_cwd = getcwd()
    virtual_layout = create_virtual_layout(files_or_pattern, h5_path, base_dir=base_dir, axis=axis)
    nx_file = NXProcessWriter(output_file, entry=output_entry, filemode=output_filemode, overwrite=overwrite)
    nx_file.write(
        virtual_layout,
        process_name,
        processing_index=processing_index,
        config=config,
        data_name=data_name,
        is_frames_stack=True,
    )
    if base_dir is not None and prev_cwd != getcwd():
        chdir(prev_cwd)


class TIFFWriter(Writer):
    def __init__(self, fname, multiframe=False, start_index=0, filemode=None, append=False, big_tiff=None):
        """
        Tiff writer.

        Parameters
        -----------
        fname: str
            Path to the output file name
        multiframe: bool, optional
            Whether to write all data in one single file. Default is False.
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            (unless multiframe is set to True).
            In this case, the output is a series of files `filename_0000.tif`,
            `filename_0001.tif`, etc. This parameter is the starting index for
            file names.
            This option is ignored when multiframe is True.
        filemode: str, optional
            DEPRECATED. Will be ignored. Please refer to 'append'
        append: bool, optional
            Whether to append data to the file rather than overwriting. Default is False.
        big_tiff: bool, optional
            Whether to write in "big tiff" format: https://www.awaresystems.be/imaging/tiff/bigtiff.html
            Default is True when multiframe is True.
            Note that default "standard" tiff cannot exceed 4 GB.

        Notes
        ------
        If multiframe is False (default), then each image will be written in a
        dedicated tiff file.
        """
        super().__init__(fname)
        self.multiframe = multiframe
        self.start_index = start_index
        self.append = append
        if big_tiff is None:
            big_tiff = multiframe
        if multiframe and not big_tiff:
            # raise error ?
            print("big_tiff was set to False while multiframe was set to True. This will probably be problematic.")
        self.big_tiff = big_tiff
        # Compat.
        self.filemode = filemode
        if filemode is not None:
            deprecation_warning("Ignored parameter 'filemode'. Please use the 'append' parameter")

    def write(self, data, *args, config=None, **kwargs):
        ext = None
        if not isinstance(data, np.ndarray):
            raise TypeError(f"data is expected to be a numpy array and not {type(data)}")
        # Single image, or multiple image in the same file
        if self.multiframe:
            volume = MultiTIFFVolume(
                self.fname,
                data=data,
                metadata={
                    "config": config,
                },
                append=self.append,
            )
            file_path = self.fname
        # Multiple image, one file per image
        else:
            if data.ndim == 2:
                data = data.reshape(1, data.shape[0], data.shape[1])
            file_path, ext = path.splitext(self.fname)
            volume = TIFFVolume(
                path.dirname(file_path),
                volume_basename=path.basename(file_path),
                data=data,
                metadata={
                    "config": config,
                },
                start_index=self.start_index,
                data_extension=ext.lstrip("."),
                overwrite=True,
            )
        volume.save()


class EDFWriter(Writer):
    def __init__(self, fname, start_index=0, filemode="w"):
        """
        EDF (ESRF Data Format) writer.

        Parameters
        -----------
        fname: str
            Path to the output file name
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            In this case, the output is a series of files `filename_0000.tif`,
            `filename_0001.edf`, etc. This parameter is the starting index for
            file names.
        """
        super().__init__(fname)
        self.filemode = filemode
        self.start_index = start_index

    def write(self, data, *args, config=None, **kwargs):
        if not isinstance(data, np.ndarray):
            raise TypeError(f"data is expected to be a numpy array and not {type(data)}")
        header = {
            "software": "nabu",
            "data": get_datetime(),
        }
        if data.ndim == 2:
            data = data.reshape(1, data.shape[0], data.shape[1])

        volume = EDFVolume(path.dirname(self.fname), data=data, start_index=self.start_index, header=header)
        volume.save()


class JP2Writer(Writer):
    def __init__(
        self,
        fname,
        start_index=0,
        filemode="wb",
        psnr=None,
        cratios=None,
        auto_convert=True,
        float_clip_values=None,
        n_threads=None,
        overwrite=False,
        single_file=True,
    ):
        """
        JPEG2000 writer. This class requires the python package `glymur` and the
        library `libopenjp2`.

        Parameters
        -----------
        fname: str
            Path to the output file name
        start_index: int, optional
            When writing a stack of images, each image is written in a dedicated file
            The output is a series of files `filename_0000.tif`, `filename_0001.tif`, etc.
            This parameter is the starting index for file names.
        psnr: list of int, optional
            The PSNR (Peak Signal-to-Noise ratio) for each jpeg2000 layer.
            This defines a quality metric for lossy compression.
            The number "0" stands for lossless compression.
        cratios: list of int, optional
            Compression ratio for each jpeg2000 layer
        auto_convert: bool, optional
            Whether to automatically cast floating point data to uint16.
            Default is True.
        float_clip_values: tuple of floats, optional
            If set to a tuple of two values (min, max), then each image values will be clipped
            to these minimum and maximum values.
        n_threads: int, optional
            Number of threads to use for encoding. Default is the number of available threads.
            Needs libopenjpeg >= 2.4.0.
        """
        super().__init__(fname)
        if not (__have_jp2k__):
            raise ValueError("Need glymur python package and libopenjp2 library")
        self.n_threads = n_threads
        # self.setup_multithread_encoding(n_threads=n_threads, what_if_not_available="ignore")
        self.filemode = filemode
        self.start_index = start_index
        self.single_file = single_file
        self.auto_convert = auto_convert
        if psnr is not None and np.isscalar(psnr):
            psnr = [psnr]
        self.psnr = psnr
        self.cratios = cratios
        self._vmin = None
        self._vmax = None
        self.overwrite = overwrite
        self.clip_float = False
        if float_clip_values is not None:
            self._float_clip_min, self._float_clip_max = float_clip_values
            self.clip_float = True

    def write(self, data, *args, **kwargs):
        if not isinstance(data, np.ndarray):
            raise TypeError(f"data is expected to be a numpy array and not {type(data)}")

        if data.ndim == 2:
            data = data.reshape(1, data.shape[0], data.shape[1])

        if self.single_file and data.ndim == 3 and data.shape[0] == 1:
            # case we will have a single file as output
            data_url = DataUrl(
                file_path=path.dirname(self.fname),
                data_path=self.fname,
                scheme=JP2KVolume.DEFAULT_DATA_SCHEME,
            )
            metadata_url = DataUrl(
                file_path=path.dirname(self.fname),
                data_path=f"{path.dirname(self.fname)}/{path.basename(self.fname)}_info.txt",
                scheme=JP2KVolume.DEFAULT_METADATA_SCHEME,
            )
            volume_basename = None
            folder = None
            extension = None
        else:
            # case we need to save it as set of file
            file_path, ext = path.splitext(self.fname)
            data_url = None
            metadata_url = None
            volume_basename = path.basename(file_path)
            folder = path.dirname(self.fname)
            extension = ext.lstrip(".")

        volume = JP2KVolume(
            folder=folder,
            start_index=self.start_index,
            cratios=self.cratios,
            psnr=self.psnr,
            n_threads=self.n_threads,
            volume_basename=volume_basename,
            data_url=data_url,
            metadata_url=metadata_url,
            data_extension=extension,
            overwrite=self.overwrite,
        )

        if data.dtype != np.uint16 and self.auto_convert:
            if self.clip_float:
                data = np.clip(data, self._float_clip_min, self._float_clip_max)
            data = rescale_data(data, 0, 65535, data_min=self._vmin, data_max=self._vmax)
            data = data.astype(np.uint16)

        volume.data = data
        config = kwargs.get("config", None)
        if config is not None:
            volume.metadata = {"config": config}
        volume.save()


class NPYWriter(Writer):
    def write(self, result, *args, **kwargs):
        np.save(self.fname, result)


class NPZWriter(Writer):
    def write(self, result, *args, **kwargs):
        save_args = {"result": result}
        config = kwargs.get("config", None)
        if config is not None:
            save_args["configuration"] = config
        np.savez(self.fname, **save_args)


class HSTVolWriter(Writer):
    """
    A writer to mimic PyHST2 ".vol" files
    """

    def __init__(self, fname, append=False, **kwargs):
        super().__init__(fname)
        self.append = append
        self._vol_writer = RawVolume(fname, overwrite=True, append=append)

    @staticmethod
    def generate_metadata(data, **kwargs):
        n_z, n_y, n_x = data.shape
        metadata = {
            "NUM_X": n_x,
            "NUM_Y": n_y,
            "NUM_Z": n_z,
            "voxelSize": kwargs.get("voxelSize", 40.0),
            "BYTEORDER": "LOWBYTEFIRST",
            "ValMin": kwargs.get("ValMin", 0.0),
            "ValMax": kwargs.get("ValMin", 1.0),
            "s1": 0.0,
            "s2": 0.0,
            "S1": 0.0,
            "S2": 0.0,
        }
        return metadata

    @staticmethod
    def sanitize_metadata(metadata):
        # To be fixed in RawVolume
        for what in ["NUM_X", "NUM_Y", "NUM_Z"]:
            metadata[what] = int(metadata[what])
        for what in ["voxelSize", "ValMin", "ValMax", "s1", "s2", "S1", "S2"]:
            metadata[what] = float(metadata[what])

    def write(self, data, *args, config=None, **kwargs):
        existing_metadata = self._vol_writer.load_metadata()
        new_metadata = self.generate_metadata(data)
        if len(existing_metadata) == 0 or not (self.append):
            # first write or append==False
            metadata = new_metadata
        else:
            # append write ; update metadata
            metadata = existing_metadata.copy()
            self.sanitize_metadata(metadata)
            metadata["NUM_Z"] += new_metadata["NUM_Z"]
        self._vol_writer.data = data
        self._vol_writer.metadata = metadata
        self._vol_writer.save()
        # Also save .xml
        self._vol_writer.save_metadata(
            url=DataUrl(
                scheme="lxml",
                file_path=self._vol_writer.metadata_url.file_path().replace(".info", ".xml"),
            )
        )


class HSTVolVolume(HSTVolWriter):
    """
    An interface to HSTVolWriter with the same API than tomoscan.esrf.volume.
    This is really not ideal, see nabu:#381
    """

    def __init__(self, **kwargs):
        file_path = kwargs.pop("file_path", None)
        if file_path is None:
            raise ValueError("Missing mandatory 'file_path' parameter")
        super().__init__(file_path, append=kwargs.pop("append", False), **kwargs)
        self.data = None
        self.metadata = None

    def save(self):
        if self.data is None:
            raise ValueError("Must set data first")
        self.write(self.data)

    def save_metadata(self):
        pass  # already done for HST part - proper metadata is not supported


# Unused - kept for compat.
Writers = {
    "h5": NXProcessWriter,
    "hdf5": NXProcessWriter,
    "nx": NXProcessWriter,
    "nexus": NXProcessWriter,
    "npy": NPYWriter,
    "npz": NPZWriter,
    "tif": TIFFWriter,
    "tiff": TIFFWriter,
    "j2k": JP2Writer,
    "jp2": JP2Writer,
    "jp2k": JP2Writer,
    "edf": EDFWriter,
    "vol": HSTVolWriter,
}



import numpy as np
from scipy.signal import find_peaks


def is_fullturn_scan(angles):
    """
    Return True if the angles correspond to a full-turn (360 degrees) scan.
    """
    angles = angles % (2 * np.pi)
    angles -= angles.min()
    sin_angles = np.sin(angles)
    min_dist = 5  # TODO find a more robust angles-based min distance, though this should cover most of the cases
    maxima = find_peaks(sin_angles, distance=min_dist)[0]
    minima = find_peaks(-sin_angles, distance=min_dist)[0]
    n_max = maxima.size
    n_min = minima.size
    # abs(n_max - n_min) actually means the following:
    #  * 0: All turns are full (eg. 2pi, 4pi)
    #  * 1: At least one half-turn remains (eg. pi, 3pi)
    if abs(n_max - n_min) == 0:
        return True
    else:
        return False


def get_halfturn_indices(angles):
    angles = angles % (2 * np.pi)
    angles -= angles.min()
    sin_angles = np.sin(angles)
    min_dist = 5  # TODO find a more robust angles-based min distance, though this should cover most of the cases
    maxima = find_peaks(sin_angles, distance=min_dist)[0]
    minima = find_peaks(-sin_angles, distance=min_dist)[0]
    extrema = np.sort(np.hstack([maxima, minima]))
    extrema -= extrema.min()
    extrema = np.hstack([extrema, [angles.size - 1]])
    res = []
    for i in range(extrema.size - 1):
        res.append((extrema[i], extrema[i + 1]))
    return res



from ..fullfield.dataset_validator import *


class HelicalDatasetValidator(FullFieldDatasetValidator):
    """Allows more freedom in the choice of the slice indices"""

    # this in the fullfield base  class is instead True
    _check_also_z = False

    def _check_slice_indices(self):
        """Slice indices can be far beyond what fullfield pipeline accepts"""
        return



import pytest
import numpy as np
from scipy.misc import ascent
from scipy.ndimage import median_filter
from nabu.testutils import generate_tests_scenarios
from nabu.cuda.utils import get_cuda_context, __has_pycuda__

if __has_pycuda__:
    from nabu.cuda.medfilt import MedianFilter
    import pycuda.gpuarray as garray


scenarios = generate_tests_scenarios(
    {
        "input_on_gpu": [False, True],
        "output_on_gpu": [False, True],
        "footprint": [(3, 3), (5, 5)],
        "mode": ["reflect", "nearest", "wrap"],
        "batched_2d": [False, True],
    }
)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    cls.data = np.ascontiguousarray(ascent()[::2, ::2][:-1, :], dtype="f")
    cls.tol = 1e-7
    cls.ctx = get_cuda_context(cleanup_at_exit=False)
    cls.allocate_numpy_arrays()
    cls.allocate_cuda_arrays()
    yield
    cls.ctx.pop()


@pytest.mark.skipif(not (__has_pycuda__), reason="Need Cuda/pycuda for this test")
@pytest.mark.usefixtures("bootstrap")
class TestMedianFilter(object):
    @classmethod
    def allocate_numpy_arrays(cls):
        shape = cls.data.shape
        cls.input = cls.data
        cls.input3d = np.tile(cls.input, (2, 1, 1))

    @classmethod
    def allocate_cuda_arrays(cls):
        shape = cls.data.shape
        cls.d_input = garray.to_gpu(cls.input)
        cls.d_output = garray.zeros_like(cls.d_input)
        cls.d_input3d = garray.to_gpu(cls.input3d)
        cls.d_output3d = garray.zeros_like(cls.d_input3d)

    # parametrize on a class method will use the same class, and launch this
    # method with different scenarios.
    @pytest.mark.parametrize("config", scenarios)
    def testMedfilt(self, config):
        if config["input_on_gpu"]:
            input_data = self.d_input if not (config["batched_2d"]) else self.d_input3d
        else:
            input_data = self.input if not (config["batched_2d"]) else self.input3d
        if config["output_on_gpu"]:
            output_data = self.d_output if not (config["batched_2d"]) else self.d_output3d
        else:
            output_data = None
        # Cuda median filter
        medfilt = MedianFilter(
            input_data.shape,
            footprint=config["footprint"],
            mode=config["mode"],
            cuda_options={"ctx": self.ctx},
        )
        res = medfilt.medfilt2(input_data, output=output_data)
        if config["output_on_gpu"]:
            res = res.get()
        # Reference (scipy)
        ref = median_filter(self.input, config["footprint"][0], mode=config["mode"])
        max_absolute_error = np.max(np.abs(res - ref))
        assert max_absolute_error < self.tol, "Something wrong with configuration %s" % str(config)



from tempfile import mkdtemp
import os
import numpy as np
import pytest
from silx.io.url import DataUrl
from silx.io import get_data
from silx.io.dictdump import dicttoh5
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.preproc.flatfield import FlatField, FlatFieldDataUrls

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.preproc.flatfield_cuda import CudaFlatFieldDataUrls, CudaFlatField


# Flats values should be O(k) so that linear interpolation between flats gives exact results
flatfield_tests_cases = {
    "simple_nearest_interp": {
        "image_shape": (100, 512),
        "radios_values": np.arange(10) + 1,
        "radios_indices": None,
        "flats_values": [0.5],
        "flats_indices": [1],
        "darks_values": [1],
        "darks_indices": [0],
        "expected_result": np.arange(0, -2 * 10, -2),
    },
    "two_flats_no_radios_indices": {
        "image_shape": (100, 512),
        "radios_values": np.arange(10) + 1,
        "radios_indices": None,
        "flats_values": [2, 11],
        "flats_indices": [0, 9],
        "darks_values": [1],
        "darks_indices": [0],
        "expected_result": np.arange(10) / (np.arange(10) + 1),
    },
    "two_flats_with_radios_indices": {
        #  Type     D   F   R   R   R   F
        #  IDX      0   1   2   3   4   5
        #  Value    1   4   9   16  25  8
        #  F_interp         5   6   7
        "image_shape": (16, 17),
        # R_k = (k + 1)**2
        "radios_values": [9, 16, 25],
        "radios_indices": [2, 3, 4],
        # F_k = k+3
        "flats_values": [4, 8],
        "flats_indices": [1, 5],
        # D_k = 1
        "darks_values": [1],
        "darks_indices": [0],
        # Expected normalization result: N_k = k
        "expected_result": [2, 3, 4],
    },
    "three_flats_srcurrent": {
        #  Type     D   F   R   R  | R   F   R   R   F
        #  IDX      0   1   2   3  | 4   5   6   7   8
        #  Value    1   4   9   16 | 25  8   46  67  14
        #  F_interp         5   6  | 7       10  12
        #  srCurrent    20  10  10 | 16  32  16  8  16
        "image_shape": (16, 17),
        "radios_values": [9, 16, 25, 46, 67],
        "radios_indices": [2, 3, 4, 6, 7],
        "flats_values": [4, 8, 14],
        "flats_indices": [1, 5, 8],
        "darks_values": [1],
        "darks_indices": [0],
        # "expected_result": [2, 3, 4, 5, 6], # without SR normalization
        "expected_result": [4, 6, 8, 10, 12],
        # sr_flat/sr_radio = 2
        "radios_srcurrent": [10, 16, 16, 16, 8],
        # "flats_srcurrent": [20, 18, 6, 4, 2],
        "flats_srcurrent": [20, 32, 16],
    },
}


def generate_test_flatfield_generalized(
    image_shape,
    radios_indices,
    radios_values,
    flats_indices,
    flats_values,
    darks_indices,
    darks_values,
    h5_fname,
    dtype=np.uint16,
):
    """
    Parameters
    -----------
    image_shape: tuple of int
        shape of each image.
    radios_indices: array of int
        Indices where radios are found in the dataset.
    radios_values: array of scalars
        Value for each radio image.
        Length must be equal to `radios_shape[0]`.
    flats_indices: array of int
        Indices where flats are found in the dataset.
    flats_values: array of scalars
        Values of flat images.
        Length must be equal to `len(flats_indices)`
    darks_indices: array of int
        Indices where darks are found in the dataset.
    darks_values: array of scalars
        Values of dark images.
        Length must be equal to `len(darks_indices)`

    Returns
    -------
    radios: numpy.ndarray
        3D array with raw radios
    darks: dict of DataUrls
        Dictionary where each key is the dark indice, and value is a DataUrl
    flats: dict of DataUrls
        Dictionary where each key is the flat indice, and value is a DataUrl
    """
    tempdir = mkdtemp(prefix="nabu_")
    testffname = os.path.join(tempdir, h5_fname)

    # Radios
    radios = np.zeros((len(radios_values),) + image_shape, dtype="f")
    n_radios = radios.shape[0]
    for i in range(n_radios):
        radios[i].fill(radios_values[i])
    img_shape = radios.shape[1:]

    # Flats
    flats = {}
    flats_urls = {}
    for i, flat_idx in enumerate(flats_indices):
        flats["flats_%06d" % flat_idx] = np.zeros(img_shape, dtype=dtype) + flats_values[i]
        flats_urls[flat_idx] = DataUrl(
            file_path=testffname, data_path=str("/flats/flats_%06d" % flat_idx), scheme="silx"
        )

    # Darks
    darks = {}
    darks_urls = {}
    for i, dark_idx in enumerate(darks_indices):
        darks["darks_%06d" % dark_idx] = np.zeros(img_shape, dtype=dtype) + darks_values[i]
        darks_urls[dark_idx] = DataUrl(
            file_path=testffname, data_path=str("/darks/darks_%06d" % dark_idx), scheme="silx"
        )

    dicttoh5(flats, testffname, h5path="/flats", mode="w")
    dicttoh5(darks, testffname, h5path="/darks", mode="a")

    return radios, flats_urls, darks_urls


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.tmp_files = []
    cls.tmp_dirs = []
    cls.n_radios = 10
    cls.n_z = 100
    cls.n_x = 512
    if __has_pycuda__:
        cls.ctx = get_cuda_context()
    yield
    # Tear-down
    for fname in cls.tmp_files:
        os.remove(fname)
    for dname in cls.tmp_dirs:
        os.rmdir(dname)


@pytest.mark.usefixtures("bootstrap")
class TestFlatField:
    def get_test_elements(self, case_name):
        config = flatfield_tests_cases[case_name]
        radios_stack, flats_url, darks_url = generate_test_flatfield_generalized(
            config["image_shape"],
            config["radios_indices"],
            config["radios_values"],
            config["flats_indices"],
            config["flats_values"],
            config["darks_indices"],
            config["darks_values"],
            "test_ff.h5",
        )
        fname = flats_url[list(flats_url.keys())[0]].file_path()
        self.tmp_files.append(fname)
        self.tmp_dirs.append(os.path.dirname(fname))
        return radios_stack, flats_url, darks_url, config

    @staticmethod
    def check_normalized_radios(radios_corr, expected_values):
        # must be the same value everywhere in the radio
        std = np.std(np.std(radios_corr, axis=-1), axis=-1)
        assert np.max(np.abs(std)) < 1e-7
        # radios values must be 0, -2, -4, ...
        assert np.allclose(radios_corr[:, 0, 0], expected_values)

    def test_flatfield_simple(self):
        """
        Test the flat-field normalization on a radios stack with 1 dark and 1 flat.

        (I - D)/(F - D)   where I = (1, 2, ...), D = 1, F = 0.5
        = (0, -2, -4, -6, ...)
        """
        radios_stack, flats_url, darks_url, config = self.get_test_elements("simple_nearest_interp")

        flatfield = FlatFieldDataUrls(radios_stack.shape, flats_url, darks_url)
        radios_corr = flatfield.normalize_radios(np.copy(radios_stack))
        self.check_normalized_radios(radios_corr, config["expected_result"])

    def test_flatfield_simple_subregion(self):
        """
        Same as test_flatfield_simple, but in a vertical subregion of the radios.
        """
        radios_stack, flats_url, darks_url, config = self.get_test_elements("simple_nearest_interp")
        end_z = 51
        radios_chunk = np.copy(radios_stack[:, :end_z, :])
        # we only have a chunk in memory. Instantiate the class with the
        # corresponding subregion to only load the relevant part of dark/flat
        flatfield = FlatFieldDataUrls(
            radios_chunk.shape,
            flats_url,
            darks_url,
            sub_region=(None, None, None, end_z),  # start_x, end_x, start_z, end_z
        )
        radios_corr = flatfield.normalize_radios(radios_chunk)
        self.check_normalized_radios(radios_corr, config["expected_result"])

    def test_flatfield_linear_interp(self):
        """
        Test flat-field normalization with 1 dark and 2 flats, with linear
        interpolation between flats.
        I   = 1   2   3   4   5   6   7   8   9   10
        D   = 1                                         (one dark)
        F   = 2                                   11    (two flats)
        F_i = 2   3   4   5   6   7   8   9   10  11   (interpolated flats)
        R   = 0  .5  .66 .75 .8  .83  .86
            = (I-D)/(F-D)
            = (I-1)/I
        """
        radios_stack, flats_url, darks_url, config = self.get_test_elements("two_flats_no_radios_indices")
        flatfield = FlatFieldDataUrls(radios_stack.shape, flats_url, darks_url)
        radios_corr = flatfield.normalize_radios(np.copy(radios_stack))
        self.check_normalized_radios(radios_corr, config["expected_result"])

        # Test 2: one of the flats is not at the beginning/end
        # I   = 1    2    3    4    5    6    7    8    9    10
        # F   = 2                       11
        # F_i = 2   3.8  5.6 7.4  9.2   11   11   11   11    11
        # R   = 0   .357  .435 .469 .488 .5  .6   .7  .8     .9
        flats_url = flats_url.copy()
        flats_url[5] = flats_url[9]
        flats_url.pop(9)
        flatfield = FlatFieldDataUrls(radios_stack.shape, flats_url, darks_url)
        radios_corr = flatfield.normalize_radios(np.copy(radios_stack))
        self.check_normalized_radios(
            radios_corr, [0.0, 0.35714286, 0.43478261, 0.46875, 0.48780488, 0.5, 0.6, 0.7, 0.8, 0.9]
        )

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        """
        Test the flat-field with cuda back-end.
        """
        radios_stack, flats_url, darks_url, config = self.get_test_elements("two_flats_no_radios_indices")
        d_radios = garray.to_gpu(radios_stack.astype("f"))
        cuda_flatfield = CudaFlatFieldDataUrls(
            d_radios.shape,
            flats_url,
            darks_url,
        )
        cuda_flatfield.normalize_radios(d_radios)
        radios_corr = d_radios.get()
        self.check_normalized_radios(radios_corr, config["expected_result"])

    # Linear interpolation, two flats, one dark
    def test_twoflats_simple(self):
        radios, flats, darks, config = self.get_test_elements("two_flats_with_radios_indices")
        FF = FlatFieldDataUrls(radios.shape, flats, darks, radios_indices=config["radios_indices"])
        FF.normalize_radios(radios)
        self.check_normalized_radios(radios, config["expected_result"])

    def _setup_numerical_issue(self):
        radios, flats, darks, config = self.get_test_elements("two_flats_with_radios_indices")

        # Retrieve the actual data for radios/darks/flats to use FlatField instead of FlatFieldDataUrl.
        # Create a setting yielding "0/0": one pixel such that flat==dark and radio==dark
        for flat_idx, flat_url in flats.items():
            flats[flat_idx] = get_data(flat_url)
            flats[flat_idx][0, 0] = 99
        for dark_idx, dark_url in darks.items():
            darks[dark_idx] = get_data(dark_url)
            darks[dark_idx][0, 0] = 99
        radios[:, 0, 0] = 99
        return radios, flats, darks, config

    def _check_numerical_issue(self, radios, expected_result, nan_value=None):
        if nan_value is None:
            assert np.alltrue(np.logical_not(np.isfinite(radios[:, 0, 0]))), "First pixel should be nan or inf"
            radios[:, 0, 0] = radios[:, 1, 1]
            self.check_normalized_radios(radios, expected_result)
        else:
            assert np.all(np.isfinite(radios)), "No inf/nan value should be there"
            assert np.allclose(radios[:, 0, 0], nan_value, atol=1e-7), (
                "Handled NaN should have nan_value=%f" % nan_value
            )
            radios[:, 0, 0] = radios[:, 1, 1]
            self.check_normalized_radios(radios, expected_result)

    def test_twoflats_numerical_issue(self):
        """
        Same as above, but for the first radio: I==Dark and Flat==Dark
        For this radio, nan is replaced with 1.
        """
        radios, flats, darks, config = self._setup_numerical_issue()
        radios0 = radios.copy()

        # FlatField without NaN handling yields NaN and raises RuntimeWarning
        FF_no_nan_handling = FlatField(
            radios.shape, flats, darks, radios_indices=config["radios_indices"], nan_value=None
        )
        with pytest.warns(RuntimeWarning):
            FF_no_nan_handling.normalize_radios(radios)
        self._check_numerical_issue(radios, config["expected_result"], None)

        # FlatField with NaN handling
        nan_value = 50
        radios = radios0.copy()
        FF_with_nan_handling = FlatField(
            radios.shape, flats, darks, radios_indices=config["radios_indices"], nan_value=nan_value
        )
        with pytest.warns(RuntimeWarning):
            FF_with_nan_handling.normalize_radios(radios)
        self._check_numerical_issue(radios, config["expected_result"], nan_value)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_twoflats_numerical_issue(self):
        """
        Same as above, with the Cuda backend
        """
        radios, flats, darks, config = self._setup_numerical_issue()
        radios0 = radios.copy()
        d_radios = garray.to_gpu(radios)
        FF_no_nan_handling = CudaFlatField(
            radios.shape, flats, darks, radios_indices=config["radios_indices"], nan_value=None
        )
        # In a cuda kernel, no one can hear you scream
        FF_no_nan_handling.normalize_radios(d_radios)
        radios = d_radios.get()
        self._check_numerical_issue(radios, config["expected_result"], None)

        # FlatField with NaN handling
        nan_value = 50
        d_radios.set(radios0)
        FF_with_nan_handling = CudaFlatField(
            radios.shape, flats, darks, radios_indices=config["radios_indices"], nan_value=nan_value
        )
        FF_with_nan_handling.normalize_radios(d_radios)
        radios = d_radios.get()
        self._check_numerical_issue(radios, config["expected_result"], nan_value)

    def test_srcurrent(self):
        radios, flats, darks, config = self.get_test_elements("three_flats_srcurrent")

        FF = FlatFieldDataUrls(
            radios.shape,
            flats,
            darks,
            radios_indices=config["radios_indices"],
            radios_srcurrent=config["radios_srcurrent"],
            flats_srcurrent=config["flats_srcurrent"],
        )
        radios_corr = FF.normalize_radios(np.copy(radios))
        self.check_normalized_radios(radios_corr, config["expected_result"])

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_srcurrent_cuda(self):
        radios, flats, darks, config = self.get_test_elements("three_flats_srcurrent")
        d_radios = garray.to_gpu(radios)

        FF = CudaFlatFieldDataUrls(
            radios.shape,
            flats,
            darks,
            radios_indices=config["radios_indices"],
            radios_srcurrent=config["radios_srcurrent"],
            flats_srcurrent=config["flats_srcurrent"],
        )
        FF.normalize_radios(d_radios)
        radios_corr = d_radios.get()
        self.check_normalized_radios(radios_corr, config["expected_result"])


# This test should be closer to the ESRF standard setting.
# There are 2 flats, one dark, 4000 radios.
#  dark :  indice=0     value=10
#  flat1 : indice=1     value=4202
#  flat2 : indice=2102  value=2101
#
# The projections have the following indices:
#  j:    0     1        1998  1999  2000  2001                   3999
# idx:  [102, 103, ..., 2100, 2101, 2203, 2204, ..., 4200, 4201, 4202]
#  Notice the gap in the middle.
#
# The linear interpolation is
#    flat_i = (n2 - i)/(n2 - n1)*flat_1  + (i - n1)/(n2 - n1)*flat_2
#  where n1 and n2 are the indices of flat_1 and flat_2 respectively.
# With the above values, we have flat_i = 4203 - i.
#
# The projections values are dark + i*(flat_i - dark),
# so that the normalization norm_i = (proj_i - dark)/(flat_i - dark) gives
#
#   idx     102    103    104    ...
#   flat    4101   4102   4103   ...
#   norm    102    103    104    ...
#


class FlatFieldTestDataset:
    # Parameters
    shp = (27, 32)
    n1 = 1  # flat indice 1
    n2 = 2102  # flat indice 2
    dark_val = 10
    darks = {0: np.zeros(shp, "f") + dark_val}
    flats = {n1: np.zeros(shp, "f") + (n2 - 1) * 2, n2: np.zeros(shp, "f") + n2 - 1}
    projs_idx = list(range(102, 2102)) + list(range(2203, 4203))  # gap in the middle

    def __init__(self):
        self._generate_projections()
        self._dump_to_h5()
        self._generate_dataurls()

    def get_flat_idx(self, proj_idx):
        flats_idx = sorted(list(self.flats.keys()))
        if proj_idx <= flats_idx[0]:
            return (flats_idx[0],)
        elif proj_idx > flats_idx[0] and proj_idx < flats_idx[1]:
            return flats_idx
        else:
            return (flats_idx[1],)

    def get_flat(self, idx):
        flatidx = self.get_flat_idx(idx)
        if len(flatidx) == 1:
            flat = self.flats[flatidx[0]]
        else:
            nf1, nf2 = flatidx
            w1 = (nf2 - idx) / (nf2 - nf1)
            flat = w1 * self.flats[nf1] + (1 - w1) * self.flats[nf2]
        return flat

    def _generate_projections(self):
        self.projs_data = np.zeros((len(self.projs_idx),) + self.shp, "f")
        self.projs = {}
        for i, proj_idx in enumerate(self.projs_idx):
            flat = self.get_flat(proj_idx)

            proj_val = self.dark_val + proj_idx * (flat[0, 0] - self.dark_val)
            self.projs[str(proj_idx)] = np.zeros(self.shp, "f") + proj_val
            self.projs_data[i] = self.projs[str(proj_idx)]

    def _dump_to_h5(self):
        self.tempdir = mkdtemp(prefix="nabu_")
        self.fname = os.path.join(self.tempdir, "projs_flats.h5")
        dicttoh5(
            {
                "projs": self.projs,
                "flats": {str(k): v for k, v in self.flats.items()},
                "darks": {str(k): v for k, v in self.darks.items()},
            },
            h5file=self.fname,
        )

    def _generate_dataurls(self):
        self.flats_urls = {}
        for idx in self.flats.keys():
            self.flats_urls[int(idx)] = DataUrl(file_path=self.fname, data_path="/flats/%d" % idx)
        self.darks_urls = {}
        for idx in self.darks.keys():
            self.darks_urls[int(idx)] = DataUrl(file_path=self.fname, data_path="/darks/0")


@pytest.fixture(scope="class")
def bootstraph5(request):
    cls = request.cls

    cls.dataset = FlatFieldTestDataset()
    n1, n2 = cls.dataset.n1, cls.dataset.n2

    # Interpolation function
    cls._weight1 = lambda i: (n2 - i) / (n2 - n1)

    cls.tol = 5e-4
    cls.tol_std = 1e-3

    yield
    # tear-down
    os.remove(cls.dataset.fname)
    os.rmdir(cls.dataset.tempdir)


@pytest.mark.usefixtures("bootstraph5")
class TestFlatFieldH5:
    def check_normalization(self, projs):
        # Check that each projection is filled with the same values
        std_projs = np.std(projs, axis=(-2, -1))
        assert np.max(np.abs(std_projs)) < self.tol_std
        # Check that the normalized radios are equal to 102, 103, 104, ...
        errs = projs[:, 0, 0] - self.dataset.projs_idx
        assert np.max(np.abs(errs)) < self.tol, "Something wrong with flat-field normalization"

    def test_flatfield(self):
        flatfield = FlatFieldDataUrls(
            self.dataset.projs_data.shape,
            self.dataset.flats_urls,
            self.dataset.darks_urls,
            radios_indices=self.dataset.projs_idx,
            interpolation="linear",
        )
        projs = np.copy(self.dataset.projs_data)
        flatfield.normalize_radios(projs)
        self.check_normalization(projs)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        d_projs = garray.to_gpu(self.dataset.projs_data)
        cuda_flatfield = CudaFlatFieldDataUrls(
            self.dataset.projs_data.shape,
            self.dataset.flats_urls,
            self.dataset.darks_urls,
            radios_indices=self.dataset.projs_idx,
        )
        cuda_flatfield.normalize_radios(d_projs)
        projs = d_projs.get()
        self.check_normalization(projs)


#
# Another test with more than two flats.
#
# Here we have
#
#   F_i = i + 2
#   R_i = i*(F_i - 1) + 1
#   N_i = (R_i - D)/(F_i - D) = i*(F_i - 1)/( F_i - 1) = i
#


def generate_test_flatfield(n_radios, radio_shape, flat_interval, h5_fname):
    radios = np.zeros((n_radios,) + radio_shape, "f")
    dark_data = np.ones(radios.shape[1:], "f")
    tempdir = mkdtemp(prefix="nabu_")
    testffname = os.path.join(tempdir, h5_fname)
    flats = {}
    flats_urls = {}
    # F_i = i + 2
    # R_i = i*(F_i - 1) + 1
    # N_i = (R_i - D)/(F_i - D) = i*(F_i - 1)/( F_i - 1) = i
    for i in range(n_radios):
        f_i = i + 2
        if (i % flat_interval) == 0:
            flats["flats_%06d" % i] = np.zeros(radio_shape, "f") + f_i
            flats_urls[i] = DataUrl(file_path=testffname, data_path=str("/flats/flats_%06d" % i), scheme="silx")
        radios[i] = i * (f_i - 1) + 1
    dark = {"dark_0000": dark_data}
    dicttoh5(flats, testffname, h5path="/flats", mode="w")
    dicttoh5(dark, testffname, h5path="/dark", mode="a")
    dark_url = {0: DataUrl(file_path=testffname, data_path="/dark/dark_0000", scheme="silx")}
    return radios, flats_urls, dark_url


@pytest.fixture(scope="class")
def bootstrap_multiflats(request):
    cls = request.cls

    n_radios = 50
    radio_shape = (20, 21)
    cls.flat_interval = 11
    h5_fname = "testff.h5"

    radios, flats, dark = generate_test_flatfield(n_radios, radio_shape, cls.flat_interval, h5_fname)
    cls.radios = radios
    cls.flats_urls = flats
    cls.darks_urls = dark
    cls.expected_results = np.arange(n_radios)

    cls.tol = 5e-4
    cls.tol_std = 1e-4

    yield
    # tear down
    os.remove(dark[0].file_path())
    os.rmdir(os.path.dirname(dark[0].file_path()))


@pytest.mark.usefixtures("bootstrap_multiflats")
class TestFlatFieldMultiFlat:
    def check_normalization(self, projs):
        # Check that each projection is filled with the same values
        std_projs = np.std(projs, axis=(-2, -1))
        assert np.max(np.abs(std_projs)) < self.tol_std
        # Check that the normalized radios are equal to 0, 1, 2, ...
        stop = (projs.shape[0] // self.flat_interval) * self.flat_interval
        errs = projs[:stop, 0, 0] - self.expected_results[:stop]
        assert np.max(np.abs(errs)) < self.tol, "Something wrong with flat-field normalization"

    def test_flatfield(self):
        flatfield = FlatFieldDataUrls(self.radios.shape, self.flats_urls, self.darks_urls, interpolation="linear")
        projs = np.copy(self.radios)
        flatfield.normalize_radios(projs)
        print(projs[:, 0, 0])
        self.check_normalization(projs)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_flatfield(self):
        d_projs = garray.to_gpu(self.radios)
        cuda_flatfield = CudaFlatFieldDataUrls(
            self.radios.shape,
            self.flats_urls,
            self.darks_urls,
        )
        cuda_flatfield.normalize_radios(d_projs)
        projs = d_projs.get()
        self.check_normalization(projs)



import pytest
import numpy as np
from nabu.testutils import get_data as nabu_get_data
from nabu.pipeline.helical.span_strategy import SpanStrategy


@pytest.fixture(scope="class")
def bootstrap_TestStrategy(request):
    cls = request.cls
    cls.abs_tol = 1.0e-6

    # from the Paul telephone dataset
    test_data = nabu_get_data("data_test_strategy.npz")

    cls.z_pix_per_proj = test_data["z_pix_per_proj"]
    cls.x_pix_per_proj = test_data["x_pix_per_proj"]
    cls.detector_shape_vh = test_data["detector_shape_vh"]
    cls.phase_margin_pix = test_data["phase_margin_pix"]
    cls.projection_angles_deg = test_data["projection_angles_deg"]
    cls.require_redundancy = test_data["require_redundancy"]
    cls.pixel_size_mm = test_data["pixel_size_mm"]

    cls.result_angle_index_span = test_data["result_angle_index_span"]
    cls.result_angles_rad = test_data["result_angles_rad"]
    cls.result_fract_complement_to_integer_shift_v = test_data["result_fract_complement_to_integer_shift_v"]
    cls.result_integer_shift_v = test_data["result_integer_shift_v"]
    cls.result_span_v = test_data["result_span_v"]
    cls.result_x_pix_per_proj = test_data["result_x_pix_per_proj"]
    cls.result_z_pix_per_proj = test_data["result_z_pix_per_proj"]

    cls.test_data = test_data


@pytest.mark.usefixtures("bootstrap_TestStrategy")
class TestStrategy:
    def test_strategy(self):
        # the python implementation is slow. so we take only a p[art of the scan
        limit = 4000
        span_info = SpanStrategy(
            z_pix_per_proj=self.z_pix_per_proj[:limit],
            x_pix_per_proj=self.x_pix_per_proj[:limit],
            detector_shape_vh=self.detector_shape_vh,
            phase_margin_pix=self.phase_margin_pix,
            projection_angles_deg=self.projection_angles_deg[:limit],
            pixel_size_mm=self.pixel_size_mm,
            require_redundancy=self.require_redundancy,
        )

        print(span_info.get_informative_string())
        chunk_info = span_info.get_chunk_info(self.result_span_v)

        for key, val in chunk_info.__dict__.items():
            reference = getattr(self, "result_" + key)
            ref_array = np.array(reference)
            res_array = np.array(val)
            if res_array.dtype in [bool, np.int32, np.int64]:
                message = f" different result for {key} attribute in the chunk_info returned value  "
                assert np.array_equal(res_array, ref_array), message
            elif res_array.dtype in [np.float32, np.float64]:
                message = f" different result for {key} attribute in the chunk_info returned value  "
                assert np.all(np.isclose(res_array, ref_array, atol=self.abs_tol)), message



import os
import numpy as np
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from tomoscan.esrf.scan.hdf5scan import HDF5TomoScan
from ..utils import check_supported, is_writeable


def get_frame_possible_urls(dataset_info, user_dir, output_dir, frame_type):
    """
    Return a list with the possible location of reduced dark/flat frames.

    Parameters
    ----------
    dataset_info: DatasetAnalyzer object
        DatasetAnalyzer object: data structure containing information on the parsed dataset
    user_dir: str or None
        User-provided directory location for the reduced frames.
    output_dir: str or None
        Output processing directory
    frame_type: str
        Frame type, can be "flats" or "darks".
    """
    check_supported(frame_type, ["flats", "darks"], "frame type")

    h5scan = dataset_info.dataset_scanner  # tomoscan object
    if frame_type == "flats":
        dataurl_default_template = h5scan.REDUCED_FLATS_DATAURLS[0]
    else:
        dataurl_default_template = h5scan.REDUCED_DARKS_DATAURLS[0]

    def make_dataurl(dirname):
        # The template formatting should be done by tomoscan in principle, but this complicates logging.
        rel_file_path = dataurl_default_template.file_path().format(
            scan_prefix=dataset_info.dataset_scanner.get_dataset_basename()
        )
        return DataUrl(
            file_path=os.path.join(dirname, rel_file_path),
            data_path=dataurl_default_template.data_path(),
            data_slice=dataurl_default_template.data_slice(),  # not sure if needed
            scheme="silx",
        )

    urls = {"user": None, "dataset": None, "output": None}

    if user_dir is not None:
        urls["user"] = make_dataurl(user_dir)

    # tomoscan.esrf.scan.hdf5scan.REDUCED_{DARKS|FLATS}_DATAURLS.file_path() is a relative path
    # Create a absolute path instead
    urls["dataset"] = make_dataurl(os.path.dirname(h5scan.master_file))

    if output_dir is not None:
        urls["output"] = make_dataurl(output_dir)

    return urls


def get_metadata_url(url, frame_type):
    """
    Return the url of the metadata stored alongside flats/darks
    """
    check_supported(frame_type, ["flats", "darks"], "frame type")
    template_url = getattr(HDF5TomoScan, "REDUCED_%s_METADATAURLS" % frame_type.upper())[0]
    return DataUrl(
        file_path=url.file_path(),
        data_path=template_url.data_path(),
        scheme="silx",
    )


def tomoscan_load_reduced_frames(dataset_info, frame_type, url):
    tomoscan_method = getattr(dataset_info.dataset_scanner, "load_reduced_%s" % frame_type)
    return tomoscan_method(
        inputs_urls=[url],
        return_as_url=True,
        return_info=True,
        metadata_input_urls=[get_metadata_url(url, frame_type)],
    )


def tomoscan_save_reduced_frames(dataset_info, frame_type, url, frames, info):
    tomoscan_method = getattr(dataset_info.dataset_scanner, "save_reduced_%s" % frame_type)
    kwargs = {"%s_infos" % frame_type: info}
    return tomoscan_method(
        frames, output_urls=[url], metadata_output_urls=[get_metadata_url(url, frame_type)], **kwargs
    )


# pylint: disable=E1136
def update_dataset_info_flats_darks(dataset_info, flatfield_mode, output_dir=None, darks_flats_dir=None):
    """
    Update a DatasetAnalyzer object with reduced flats/darks (hereafter "reduced frames").

    How the reduced frames are loaded/computed/saved will depend on the "flatfield_mode" parameter.

    The principle is the following:
    (1) Attempt at loading already-computed reduced frames (XXX_darks.h5 and XXX_flats.h5):
       - First check files in the user-defined directory 'darks_flats_dir'
       - Then try to load from files located alongside the .nx dataset (dataset directory)
       - Then try to load from output_dir, if provided
    (2) If loading fails, or flatfield_mode == "force_compute", compute the reduced frames.
    (3) Save these reduced frames
       - Save in darks_flats_dir, if provided by user
       - Otherwise, save in the data directory (next to the .nx file), if write access OK
       - Otherwise, save in output directory
    """
    if flatfield_mode is False:
        return
    logger = dataset_info.logger
    frames_types = ["darks", "flats"]

    reduced_frames_urls = {}
    for frame_type in frames_types:
        reduced_frames_urls[frame_type] = get_frame_possible_urls(dataset_info, darks_flats_dir, output_dir, frame_type)

    reduced_frames = dict.fromkeys(frames_types, None)

    #
    # Try to load frames
    #
    def load_reduced_frame(url, frame_type, frames_loaded, reduced_frames):
        if frames_loaded[frame_type]:
            return
        frames, info = tomoscan_load_reduced_frames(dataset_info, frame_type, url)
        if frames not in (None, {}):
            dataset_info.logger.info("Loaded %s from %s" % (frame_type, url.file_path()))
            frames_loaded[frame_type] = True
            reduced_frames[frame_type] = frames, info
        else:
            msg = "Could not load %s from %s" % (frame_type, url.file_path())
            logger.error(msg)

    frames_loaded = dict.fromkeys(frames_types, False)
    if flatfield_mode != "force-compute":
        for load_from in ["user", "dataset", "output"]:  # in that order
            for frame_type in frames_types:
                url = reduced_frames_urls[frame_type][load_from]
                if url is None:
                    continue  # cannot load from this source (eg. undefined folder)
                load_reduced_frame(url, frame_type, frames_loaded, reduced_frames)
            if all(frames_loaded.values()):
                break

    if not all(frames_loaded.values()) and flatfield_mode == "force-load":
        raise ValueError("Could not load darks/flats (using 'force-load')")

    #
    # COMPAT. Keep DataUrl - won't be needed in future versions when pipeline will use FlatField
    # instead of FlatFieldDataUrl
    frames_urls = reduced_frames.copy()
    #

    # Compute reduced frames, if needed
    #
    if reduced_frames["flats"] is None:
        reduced_frames["flats"] = dataset_info.dataset_scanner.compute_reduced_flats(return_info=True)
    if reduced_frames["darks"] is None:
        reduced_frames["darks"] = dataset_info.dataset_scanner.compute_reduced_darks(return_info=True)

    if reduced_frames["darks"][0] == {} or reduced_frames["flats"][0] == {}:
        raise ValueError(
            "Could not get any reduced flat/dark. This probably means that no already-reduced flats/darks were found and that the dataset itself does not have any flat/dark"
        )

    #
    # Save reduced frames
    #

    def save_reduced_frame(url, frame_type, frames_saved):
        frames, info = reduced_frames[frame_type]
        tomoscan_save_reduced_frames(dataset_info, frame_type, url, frames, info)
        dataset_info.logger.info("Saved reduced %s to %s" % (frame_type, url.file_path()))
        frames_saved[frame_type] = True

    frames_saved = dict.fromkeys(frames_types, False)
    if not all(frames_loaded.values()):
        for save_to in ["user", "dataset", "output"]:  # in that order
            for frame_type in frames_types:
                if frames_loaded[frame_type]:
                    continue  # already loaded
                url = reduced_frames_urls[frame_type][save_to]
                if url is None:
                    continue  # cannot load from this source (eg. undefined folder)
                if not is_writeable(os.path.dirname(url.file_path())):
                    continue
                save_reduced_frame(url, frame_type, frames_saved)
                # COMPAT.
                if frames_urls[frame_type] is None:
                    frames_urls[frame_type] = tomoscan_load_reduced_frames(dataset_info, frame_type, url)
                #
            if all(frames_saved.values()):
                break

    dataset_info.flats = frames_urls["flats"][0]  # reduced_frames["flats"] # in future versions
    dataset_info.flats_srcurrent = frames_urls["flats"][1].machine_electric_current
    # This is an extra check to avoid having more than 1 (reduced) dark.
    # FlatField only works with exactly 1 (reduced) dark (having more than 1 series of darks makes little sense)
    # This is normally prevented by tomoscan HDF5FramesReducer, but let's add this extra check
    darks_ = frames_urls["darks"][0]  # reduced_frames["darks"] # in future versions
    if len(darks_) > 1:
        dark_idx = sorted(darks_.keys())[0]
        dataset_info.logger.error("Found more that one series of darks. Keeping only the first one")
        darks_ = {dark_idx: darks_[dark_idx]}
    #
    dataset_info.darks = darks_
    dataset_info.darks_srcurrent = frames_urls["darks"][1].machine_electric_current


# tomoscan "compute_reduced_XX" is quite slow. If needed, here is an alternative implementation
def my_reduce_flats(di):
    res = {}
    with HDF5File(di.dataset_hdf5_url.file_path(), "r") as f:
        for data_slice in di.get_data_slices("flats"):
            data = f[di.dataset_hdf5_url.data_path()][data_slice.start : data_slice.stop]
            res[data_slice.start] = np.median(data, axis=0)
    return res
